from __future__ import annotations

from typing import Optional, List, TYPE_CHECKING

from analogai.foundation.core.value_objects.attribute import Attribute

if TYPE_CHECKING:
    from analogai.foundation.modules.notion.notion import Notion
    from analogai.foundation.common.dtos.notion_expression import NotionExpression


def _format_attribute(attribute: Attribute) -> Optional[str]:
    if attribute is None:
        return None
    tag = attribute.tag
    value = attribute.value
    unit = attribute.unit
    if tag is None and value is None:
        return None
    if value is None:
        return str(tag)
    if unit:
        return f"{tag} {value} {unit}"
    return f"{tag} {value}"


def serialize_notion(
    caption: str,
    attributes: Optional[List[Attribute]] = None,
) -> str:
    caption = caption.strip() if caption else ""
    if not attributes:
        return caption
    parts: List[str] = []
    for attribute in attributes:
        formatted = _format_attribute(attribute)
        if formatted:
            parts.append(formatted)
    if not parts:
        return caption
    return f"{caption} has {', '.join(parts)}"


def serialize_notion_entity(notion: "Notion") -> str:
    return serialize_notion(notion.caption, notion.attributes)


def serialize_notion_expression(expression: "NotionExpression") -> str:
    return serialize_notion(expression.caption, expression.attributes)


__all__ = [
    "serialize_notion",
    "serialize_notion_entity",
    "serialize_notion_expression",
]



from enum import Enum


class Certainty(Enum):
    CERTAIN = "certain"
    UNCERTAIN = "uncertain"
    NEGATED = "negated"

    def to_probability(self) -> float:
        if self is Certainty.CERTAIN:
            return 1.0
        if self is Certainty.UNCERTAIN:
            return 0.5
        return 0.0


def certainty_from_value(value: "Certainty | str | None") -> "Certainty":
    if value is None:
        return Certainty.CERTAIN
    if isinstance(value, Certainty):
        return value
    try:
        return Certainty(str(value))
    except ValueError:
        return Certainty.CERTAIN



from abc import ABC, abstractmethod
from typing import List

class BaseChainHandler(ABC):
    def __init__(self):
        self._next_handler = None

    def set_next(self, handler):
        self._next_handler = handler
        return handler
    
    @abstractmethod
    def execute_chain_handler(self, *args, **kwargs):
        pass
    
    def process(self, *args, **kwargs):
        pass

    @abstractmethod
    def execute(self, *args, **kwargs):
        pass

class ChainHandler(BaseChainHandler):
    def execute_chain_handler(self, *args, **kwargs):
        result = self.execute(*args, **kwargs)
        if result is not None:
            return result
        if self._next_handler:
            return self._next_handler.execute_chain_handler(*args, **kwargs)
        return None

class ChainCreator:
    def __init__(self):
        pass

    @staticmethod
    def create(chain_items: List[ChainHandler]):
        if not chain_items:
            raise ValueError("Chain items list cannot be empty")
        for i in range(len(chain_items) - 1):
            chain_items[i].set_next(chain_items[i + 1])
        return _ChainHeadAdapter(chain_items[0])

class _ChainHeadAdapter:
    def __init__(self, head: ChainHandler):
        self._head = head

    def execute(self, *args, **kwargs):
        return self._head.execute_chain_handler(*args, **kwargs)

    def __getattr__(self, name):
        return getattr(self._head, name)





import asyncio
import unittest
from types import SimpleNamespace
from unittest.mock import Mock, patch
from fastapi import Request
from analogai.foundation.modules.agent.agent import Agent, Role
from analogai.foundation.modules.user.user import User
from analogai.foundation.tests.functional.services.setup import ServicesTestConfig
from analogai.foundation.extensions.authorization.authorization_factory import AuthorizationFactory
from analogai.foundation.extensions.authorization.core.enums.operation import AuthorizationOperation
from analogai.foundation.extensions.authorization.models import AuthorizationContext
import uuid

class TestAuthorizationFunctional(unittest.TestCase):
    def setUp(self):
        self.config = ServicesTestConfig()
        self.user_service = self.config.user_service
        self.agent_service = self.config.agent_service
        self.user = self.config.user
        self.agent = self.config.agent

    def tearDown(self):
        self.config.tear_down()

    @patch("analogai.foundation.extensions.authorization.authorization_factory.AuthenticationFactory")
    def test_authorization_context_authority_with_role(self, authentication_factory_cls):
        auth_factory = authentication_factory_cls.return_value
        
        test_authority = 0.85
        test_user = User(id=str(uuid.uuid4()), name="Test User", email=f"test_role_{uuid.uuid4()}@example.com")
        self.user_service.create(test_user.id, test_user.name, test_user.email)
        
        test_agent = self.agent_service.create(
            creator=test_user,
            id=str(uuid.uuid4()),
            name="Test Agent",
            knowledge_base="Test KB",
            roles=[Role(user=test_user, authority=test_authority)]
        )

        async def auth_dependency(request: Request) -> User:
            return test_user

        auth_factory.create_authentication_dependency.return_value = auth_dependency

        factory = AuthorizationFactory()
        dependency = factory.create_authorization_dependency()

        request = Mock(spec=Request)
        request.headers = {"Authorization": "Bearer auth-token"}

        context = asyncio.run(dependency(request, agent_id=test_agent.id))

        self.assertIsInstance(context, AuthorizationContext)
        self.assertEqual(context.user.id, test_user.id)
        self.assertEqual(context.agent.id, test_agent.id)
        
        user_role = next(role for role in test_agent.roles if role.user.id == test_user.id)
        self.assertEqual(context.authority, user_role.authority)
        self.assertEqual(context.authority, test_authority)
        self.assertEqual(context.operation, AuthorizationOperation.CHAT)

    @patch("analogai.foundation.extensions.authorization.authorization_factory.AuthenticationFactory")
    def test_authorization_context_authority_without_role(self, authentication_factory_cls):
        auth_factory = authentication_factory_cls.return_value
        
        user_without_role = User(id=str(uuid.uuid4()), name="User Without Role", email="no-role@example.com")
        self.user_service.create(user_without_role.id, user_without_role.name, user_without_role.email)
        
        test_authority = 0.7
        test_agent = self.agent_service.create(
            creator=self.user,
            id=str(uuid.uuid4()),
            name="Test Agent",
            knowledge_base="Test KB",
            roles=[Role(user=self.user, authority=test_authority)]
        )

        async def auth_dependency(request: Request) -> User:
            return user_without_role

        auth_factory.create_authentication_dependency.return_value = auth_dependency

        factory = AuthorizationFactory()
        dependency = factory.create_authorization_dependency()

        request = Mock(spec=Request)
        request.headers = {"Authorization": "Bearer auth-token"}

        context = asyncio.run(dependency(request, agent_id=test_agent.id))

        self.assertIsInstance(context, AuthorizationContext)
        self.assertEqual(context.user.id, user_without_role.id)
        self.assertEqual(context.agent.id, test_agent.id)
        self.assertEqual(context.authority, 0.0)
        self.assertEqual(context.operation, AuthorizationOperation.CHAT)

    @patch("analogai.foundation.extensions.authorization.authorization_factory.AuthenticationFactory")
    def test_authority_extraction_from_persisted_agent(self, authentication_factory_cls):
        auth_factory = authentication_factory_cls.return_value
        
        test_authority = 0.65
        test_user = User(id=str(uuid.uuid4()), name="Test User", email=f"test_persisted_{uuid.uuid4()}@example.com")
        self.user_service.create(test_user.id, test_user.name, test_user.email)
        
        test_agent = self.agent_service.create(
            creator=test_user,
            id=str(uuid.uuid4()),
            name="Test Agent",
            knowledge_base="Test KB",
            roles=[Role(user=test_user, authority=test_authority)]
        )
        
        retrieved_agent = self.agent_service.find(test_agent.id)
        self.assertIsNotNone(retrieved_agent)
        self.assertEqual(len(retrieved_agent.roles), 1)
        self.assertEqual(retrieved_agent.roles[0].authority, test_authority)

        async def auth_dependency(request: Request) -> User:
            return test_user

        auth_factory.create_authentication_dependency.return_value = auth_dependency

        factory = AuthorizationFactory()
        dependency = factory.create_authorization_dependency()

        request = Mock(spec=Request)
        request.headers = {"Authorization": "Bearer auth-token"}

        context = asyncio.run(dependency(request, agent_id=test_agent.id))

        self.assertIsInstance(context, AuthorizationContext)
        retrieved_role = next(role for role in retrieved_agent.roles if role.user.id == test_user.id)
        self.assertEqual(context.authority, retrieved_role.authority)
        self.assertEqual(context.authority, test_authority)

    @patch("analogai.foundation.extensions.authorization.authorization_factory.AuthenticationFactory")
    def test_authorization_context_authority_guest_api_token(self, authentication_factory_cls):
        auth_factory = authentication_factory_cls.return_value

        test_user = User(id=str(uuid.uuid4()), name="Guest User", email=f"guest_{uuid.uuid4()}@example.com")
        self.user_service.create(test_user.id, test_user.name, test_user.email)

        test_agent = self.agent_service.create(
            creator=test_user,
            id=str(uuid.uuid4()),
            name="Guest Agent",
            knowledge_base="Guest KB",
        )

        async def auth_dependency(request: Request) -> User:
            return test_user

        auth_factory.create_authentication_dependency.return_value = auth_dependency

        factory = AuthorizationFactory()
        dependency = factory.create_authorization_dependency()

        request = Mock(spec=Request)
        request.headers = {"Authorization": "Bearer auth-token"}
        request.state = SimpleNamespace(
            api_auth_type="guest",
            api_agent_id=test_agent.id,
            api_user_email=test_user.email,
        )

        context = asyncio.run(dependency(request, agent_id=test_agent.id))

        self.assertIsInstance(context, AuthorizationContext)
        self.assertEqual(context.user.id, test_user.id)
        self.assertEqual(context.agent.id, test_agent.id)
        self.assertEqual(context.authority, 0.3)
        self.assertEqual(context.operation, AuthorizationOperation.CHAT)

if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import Mock
from analogai.foundation.core.observables.statement_observable_service import StatementObservableService, StatementObserver
from analogai.foundation.modules.direct_statement.direct_statement import DirectStatement as Statement
from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.modules.user.user import User
from analogai.foundation.core.value_objects.relation import Relation


class TestStatementObserver(StatementObserver):
    def __init__(self):
        self.created_statements = []
        self.updated_statements = []
    
    def on_statement_created(self, statement):
        self.created_statements.append(statement)
    
    def on_statement_updated(self, statement):
        self.updated_statements.append(statement)


class TestStatementObservableService(unittest.TestCase):
    def setUp(self):
        self.service = StatementObservableService()
        self.user = User(id="user1", name="Test User", email="test@example.com")
        self.notion1 = Notion(user_id="user1", agent_id="agent1", caption="Subject", id="notion1")
        self.notion2 = Notion(user_id="user1", agent_id="agent1", caption="Object", id="notion2")
        self.statement = Statement(
            user_id="user1",
            agent_id="agent1",
            subject_notion=self.notion1,
            relation=Relation.from_string("is"),
            complement_notion=self.notion2,
            probability=0.8,
            validity=0.9
        )
    
    def test_add_observer(self):
        observer = Mock(spec=StatementObserver)
        self.service.add_observer(observer)
        
        self.assertIn(observer, self.service._observers)
    
    def test_remove_observer(self):
        observer = Mock(spec=StatementObserver)
        self.service.add_observer(observer)
        self.service.remove_observer(observer)
        
        self.assertNotIn(observer, self.service._observers)
    
    def test_notify_statement_created_calls_observer(self):
        observer = Mock(spec=StatementObserver)
        observer.on_statement_created = Mock()
        self.service.add_observer(observer)
        
        self.service.notify_statement_created(self.statement)
        
        observer.on_statement_created.assert_called_once_with(self.statement)
    
    def test_notify_statement_updated_calls_observer(self):
        observer = Mock(spec=StatementObserver)
        observer.on_statement_updated = Mock()
        self.service.add_observer(observer)
        
        self.service.notify_statement_updated(self.statement)
        
        observer.on_statement_updated.assert_called_once_with(self.statement)
    
    def test_notify_statement_created_calls_all_observers(self):
        observer1 = Mock(spec=StatementObserver)
        observer1.on_statement_created = Mock()
        observer2 = Mock(spec=StatementObserver)
        observer2.on_statement_created = Mock()
        
        self.service.add_observer(observer1)
        self.service.add_observer(observer2)
        
        self.service.notify_statement_created(self.statement)
        
        observer1.on_statement_created.assert_called_once_with(self.statement)
        observer2.on_statement_created.assert_called_once_with(self.statement)
    
    def test_notify_statement_updated_calls_all_observers(self):
        observer1 = Mock(spec=StatementObserver)
        observer1.on_statement_updated = Mock()
        observer2 = Mock(spec=StatementObserver)
        observer2.on_statement_updated = Mock()
        
        self.service.add_observer(observer1)
        self.service.add_observer(observer2)
        
        self.service.notify_statement_updated(self.statement)
        
        observer1.on_statement_updated.assert_called_once_with(self.statement)
        observer2.on_statement_updated.assert_called_once_with(self.statement)
    
    def test_notify_statement_created_handles_missing_method_gracefully(self):
        observer = Mock()
        del observer.on_statement_created
        
        self.service.add_observer(observer)
        
        self.service.notify_statement_created(self.statement)
    
    def test_notify_statement_updated_handles_missing_method_gracefully(self):
        observer = Mock()
        del observer.on_statement_updated
        
        self.service.add_observer(observer)
        
        self.service.notify_statement_updated(self.statement)
    
    def test_notify_statement_created_with_concrete_observer(self):
        observer = TestStatementObserver()
        self.service.add_observer(observer)
        
        self.service.notify_statement_created(self.statement)
        
        self.assertEqual(len(observer.created_statements), 1)
        self.assertEqual(observer.created_statements[0], self.statement)
        self.assertEqual(len(observer.updated_statements), 0)
    
    def test_notify_statement_updated_with_concrete_observer(self):
        observer = TestStatementObserver()
        self.service.add_observer(observer)
        
        self.service.notify_statement_updated(self.statement)
        
        self.assertEqual(len(observer.updated_statements), 1)
        self.assertEqual(observer.updated_statements[0], self.statement)
        self.assertEqual(len(observer.created_statements), 0)
    
    def test_notify_both_created_and_updated(self):
        observer = TestStatementObserver()
        self.service.add_observer(observer)
        
        statement1 = Statement(
            user_id="user1",
            agent_id="agent1",
            subject_notion=self.notion1,
            relation=Relation.from_string("is"),
            complement_notion=self.notion2,
            probability=0.8,
            validity=0.9
        )
        statement2 = Statement(
            user_id="user1",
            agent_id="agent1",
            subject_notion=self.notion1,
            relation=Relation.from_string("has"),
            complement_notion=self.notion2,
            probability=0.7,
            validity=0.8
        )
        
        self.service.notify_statement_created(statement1)
        self.service.notify_statement_updated(statement2)
        
        self.assertEqual(len(observer.created_statements), 1)
        self.assertEqual(observer.created_statements[0], statement1)
        self.assertEqual(len(observer.updated_statements), 1)
        self.assertEqual(observer.updated_statements[0], statement2)


if __name__ == '__main__':
    unittest.main()



from typing import Optional, List
from analogai.foundation.extensions.fallback.repositories.fallback_repository import FallbackRepository
from analogai.foundation.extensions.fallback import Fallback
from analogai.foundation.infrastructure.storage.storage_adapter import StorageAdapter
from analogai.foundation.infrastructure.storage.mongodb.mongodb_storage_adapter_impl import MongoDBStorageAdapterImpl
from analogai.foundation.extensions.fallback.mappers.fallback_storage_mapper import FallbackStorageMapper


class MongoDBFallbackRepositoryImpl(FallbackRepository):
    COLLECTION_NAME = "fallbacks"

    def __init__(self, storage_adapter: Optional[StorageAdapter] = None):
        self.storage_adapter = storage_adapter if storage_adapter is not None else MongoDBStorageAdapterImpl()

    def create(self, fallback: Fallback) -> Fallback:
        data = FallbackStorageMapper.to_dict(fallback)
        stored_data = self.storage_adapter.store(self.COLLECTION_NAME, data)
        if not stored_data:
            raise ValueError(f"Failed to store fallback {fallback.id}")
        return fallback

    def find_by_id(self, fallback_id: str) -> Optional[Fallback]:
        data = self.storage_adapter.retrieve_by_id(self.COLLECTION_NAME, fallback_id)
        if not data:
            return None
        return FallbackStorageMapper.from_dict(data)

    def find_by_agent_id(self, agent_id: str) -> List[Fallback]:
        fallbacks_data = self.storage_adapter.retrieve_by_attributes(
            self.COLLECTION_NAME,
            {"agent_id": agent_id}
        )
        return [FallbackStorageMapper.from_dict(data) for data in fallbacks_data]



from typing import List, Any


class ObservableService:
    def __init__(self):
        self._observers: List[Any] = []

    def add_observer(self, observer: Any) -> None:
        if observer not in self._observers:
            self._observers.append(observer)

    def remove_observer(self, observer: Any) -> None:
        if observer in self._observers:
            self._observers.remove(observer)


import unittest
from unittest.mock import MagicMock
from analogai.foundation.modules.user.user_service import UserService
from analogai.foundation.modules.user.user_repository import UserRepository
from analogai.foundation.modules.user.user import User


test_user_id = "test_user_id"
test_authority = 1.0


class TestUserService(unittest.TestCase):
    def setUp(self):
        self.mock_user_repository = MagicMock(spec=UserRepository)
        self.test_user = User(id=test_user_id, name="Test User", email="test@example.com")
        self.user_service = UserService(self.mock_user_repository)

    def test_find_user_success(self):
        self.mock_user_repository.find_by_id.return_value = self.test_user
        result = self.user_service.find(test_user_id)
        self.mock_user_repository.find_by_id.assert_called_once_with(test_user_id)
        self.assertEqual(result, self.test_user)

    def test_find_user_not_found(self):
        self.mock_user_repository.find_by_id.return_value = None
        user = self.user_service.find(test_user_id)
        self.assertIsNone(user)
        self.mock_user_repository.find_by_id.assert_called_once_with(test_user_id)

    def test_find_by_email_success(self):
        test_email = "test@example.com"
        self.mock_user_repository.find_by_email.return_value = self.test_user
        result = self.user_service.find_by_email(test_email)
        self.mock_user_repository.find_by_email.assert_called_once_with(test_email)
        self.assertEqual(result, self.test_user)

    def test_find_by_email_not_found(self):
        test_email = "nonexistent@example.com"
        self.mock_user_repository.find_by_email.return_value = None
        user = self.user_service.find_by_email(test_email)
        self.assertIsNone(user)
        self.mock_user_repository.find_by_email.assert_called_once_with(test_email)

    def test_update_user_success(self):
        updated_user = User(id=test_user_id, name="Updated User", email="updated@example.com")
        self.mock_user_repository.find_by_id.return_value = self.test_user
        self.mock_user_repository.update.return_value = updated_user

        result = self.user_service.update(updated_user)

        self.mock_user_repository.find_by_id.assert_called_once_with(test_user_id)
        self.mock_user_repository.update.assert_called_once_with(updated_user)
        self.assertEqual(result, updated_user)

    def test_update_user_not_found(self):
        updated_user = User(id="nonexistent_id", name="Updated User", email="updated@example.com")
        self.mock_user_repository.find_by_id.return_value = None

        with self.assertRaises(ValueError) as context:
            self.user_service.update(updated_user)

        self.assertEqual(str(context.exception), f"User with ID {updated_user.id} not found")
        self.mock_user_repository.find_by_id.assert_called_once_with(updated_user.id)
        self.mock_user_repository.update.assert_not_called()

    def test_create_user_success(self):
        new_user = User(id=test_user_id, name="New User", email="new@example.com")
        self.mock_user_repository.find_by_email.return_value = None
        self.mock_user_repository.create.return_value = new_user

        result = self.user_service.create(test_user_id, "New User", "new@example.com")

        self.mock_user_repository.find_by_email.assert_called_once_with("new@example.com")
        self.mock_user_repository.create.assert_called_once()
        self.assertEqual(result, new_user)

    def test_create_user_with_duplicate_email_raises_error(self):
        existing_user = User(id="existing_id", name="Existing User", email="duplicate@example.com")
        self.mock_user_repository.find_by_email.return_value = existing_user

        with self.assertRaises(ValueError) as context:
            self.user_service.create(test_user_id, "New User", "duplicate@example.com")

        self.assertIn("already exists", str(context.exception))
        self.assertIn("duplicate@example.com", str(context.exception))
        self.mock_user_repository.find_by_email.assert_called_once_with("duplicate@example.com")
        self.mock_user_repository.create.assert_not_called()

    def test_create_user_with_timezone(self):
        timezone = "America/New_York"
        new_user = User(id=test_user_id, name="New User", email="new@example.com", timezone=timezone)
        self.mock_user_repository.find_by_email.return_value = None
        self.mock_user_repository.create.return_value = new_user

        result = self.user_service.create(test_user_id, "New User", "new@example.com", timezone=timezone)

        self.mock_user_repository.find_by_email.assert_called_once_with("new@example.com")
        self.mock_user_repository.create.assert_called_once()
        self.assertEqual(result, new_user)
        self.assertEqual(result.timezone, timezone)

    def test_user_has_default_timezone(self):
        new_user = User(id=test_user_id, name="New User", email="new@example.com")
        self.assertIsNotNone(new_user.timezone)

    def test_user_with_nationality_and_location(self):
        user = User(
            id=test_user_id,
            name="Test User",
            email="test@example.com",
            nationality="US",
            location="New York"
        )
        self.assertEqual(user.nationality, "US")
        self.assertEqual(user.location, "New York")

    def test_delete_user_success(self):
        self.mock_user_repository.delete.return_value = True

        result = self.user_service.delete(test_user_id)

        self.mock_user_repository.delete.assert_called_once_with(test_user_id)
        self.assertTrue(result)


if __name__ == '__main__':
    unittest.main()



import json
from typing import List, Optional

from analogai.foundation.modules.conditionals.hypothetical_statement import HypotheticalStatement
from analogai.foundation.modules.conditionals.hypothetical_statement_repository import HypotheticalStatementRepository
from analogai.foundation.infrastructure.storage.neo4j.neo4j_gateway import Neo4jGateway


class Neo4jHypotheticalStatementRepositoryImpl(HypotheticalStatementRepository):
    LABEL = "hypothetical_statement"

    def __init__(self):
        self.neo4j_gateway = Neo4jGateway()

    def create(self, hypothetical_statement: HypotheticalStatement) -> HypotheticalStatement:
        causes_serialized = self._serialize_causes(hypothetical_statement.causes)
        with self.neo4j_gateway.session() as session:
            session.run(
                f"CREATE (c:{self.LABEL} {{id: $id, agent_id: $agent_id, effect_belief_id: $effect_belief_id, "
                "probability: $probability, causes: $causes})",
                id=hypothetical_statement.id,
                agent_id=hypothetical_statement.agent_id,
                effect_belief_id=hypothetical_statement.effect_belief_id,
                probability=hypothetical_statement.probability,
                causes=causes_serialized,
            )
        return hypothetical_statement

    def find_by_id(self, hypothetical_statement_id: str) -> Optional[HypotheticalStatement]:
        with self.neo4j_gateway.session() as session:
            result = session.run(
                f"MATCH (c:{self.LABEL} {{id: $id}}) RETURN c",
                id=hypothetical_statement_id,
            )
            record = result.single()
            if not record:
                return None
            return self._from_node(record["c"])

    def find_by_agent_id(self, agent_id: str) -> List[HypotheticalStatement]:
        with self.neo4j_gateway.session() as session:
            result = session.run(
                f"MATCH (c:{self.LABEL} {{agent_id: $agent_id}}) RETURN c",
                agent_id=agent_id,
            )
            return [self._from_node(record["c"]) for record in result]

    def _serialize_causes(self, causes) -> str:
        serialized = []
        for c in causes:
            clause = {}
            for k, v in c.criteria.items():
                try:
                    key = k.value
                except AttributeError:
                    continue
                try:
                    clause[key] = v.value
                except AttributeError:
                    clause[key] = v
            serialized.append(clause)
        return json.dumps(serialized)

    def _from_node(self, node) -> HypotheticalStatement:
        from analogai.foundation.core.value_objects.cause import Cause

        causes_raw = json.loads(node.get("causes", "[]"))
        causes = [Cause(self._deserialize_criteria(raw or {})) for raw in causes_raw]
        return HypotheticalStatement(
            id=node.get("id"),
            agent_id=node["agent_id"],
            causes=causes,
            effect_belief_id=node.get("effect_belief_id", ""),
            probability=node.get("probability", 1.0),
        )

    def _deserialize_criteria(self, raw: dict) -> dict:
        from analogai.foundation.common.enums.criterion import Criterion

        criteria = {}
        for key, value in raw.items():
            try:
                criterion = Criterion(key)
            except ValueError:
                continue
            if criterion == Criterion.Relation:
                criteria[criterion] = str(value) if value is not None else value
            else:
                if type(value) is list:
                    criteria[criterion] = tuple(value)
                else:
                    criteria[criterion] = value
        return criteria



from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Callable
from analogai.foundation.modules.user.user import User
from analogai.foundation.modules.agent.agent import Agent
from analogai.foundation.extensions.authorization.core.services.authorization_service import AuthorizationService
from analogai.foundation.extensions.authorization.core.enums.operation import AuthorizationOperation
from analogai.foundation.common.logging.app_logger import app_logger

class AuthorizationErrorReason(Enum):
    NOT_FOUND = "not_found"
    ACCESS_DENIED = "access_denied"
    INVALID_REQUEST = "invalid_request"
    INTERNAL_ERROR = "internal_error"

class AuthorizationGuardError(Exception):
    def __init__(self, message: str, reason: AuthorizationErrorReason):
        super().__init__(message)
        self.reason = reason

@dataclass(frozen=True)
class ErrorReasonRule:
    predicate: Callable[[str], bool]
    reason: AuthorizationErrorReason

class AuthorizationErrorReasonResolver:
    def __init__(self, rules: list[ErrorReasonRule] | None = None):
        self._rules = rules or [
            ErrorReasonRule(
                predicate=lambda message: "not found" in message,
                reason=AuthorizationErrorReason.NOT_FOUND,
            ),
        ]
        self._default_reason = AuthorizationErrorReason.ACCESS_DENIED

    def resolve(self, message: str) -> AuthorizationErrorReason:
        lowered = message.lower()
        for rule in self._rules:
            if rule.predicate(lowered):
                return rule.reason
        return self._default_reason

class AuthorizationGuard(ABC):

    def __init__(self, authorization_service: AuthorizationService):
        self.authorization_service = authorization_service
        self._error_reason_resolver = AuthorizationErrorReasonResolver()
    
    def authorize_agent(
        self,
        agent_id: str,
        current_user: User,
        operation: AuthorizationOperation = AuthorizationOperation.CHAT
    ) -> Agent:

        if not current_user:
            app_logger.error("Authorization failed: No authenticated user")
            self._raise_error("User not authenticated", AuthorizationErrorReason.INVALID_REQUEST)

        try:
            agent = self.authorization_service.find_and_authorize_agent(
                current_user,
                agent_id,
                operation,
            )
            if operation == AuthorizationOperation.EDIT and not self._has_edit_access(current_user, agent):
                message = f"User {current_user.id} does not have edit access to agent {agent_id}"
                app_logger.error(f"Authorization failed: {message}")
                self._raise_error(message, AuthorizationErrorReason.ACCESS_DENIED)

            app_logger.info(f"Authorized user {current_user.id} for agent {agent.id}")
            return agent

        except ValueError as error:
            message = str(error)
            app_logger.error(f"Authorization failed: {message}")
            self._raise_error(message, self._error_reason_resolver.resolve(message))

        except Exception as error:
            message = str(error)
            app_logger.error(f"Authorization failed with unexpected error: {message}")
            self._raise_error("Authorization failed", AuthorizationErrorReason.INTERNAL_ERROR)

    def _has_edit_access(self, current_user: User, agent: Agent) -> bool:
        return any(role.user and role.user.id == current_user.id for role in agent.roles or [])

    def _raise_error(self, message: str, reason: AuthorizationErrorReason) -> None:
        raise AuthorizationGuardError(message, reason)

    @abstractmethod
    async def get_accessible_agent(
        self,
        agent_id: str,
        current_user: User = None,
        operation: AuthorizationOperation = AuthorizationOperation.CHAT,
        **context
    ) -> Agent:

        pass




from typing import Any, List, Optional, Dict
from pymongo import MongoClient
from pymongo.errors import OperationFailure, ServerSelectionTimeoutError, ConnectionFailure, ConfigurationError
from analogai.foundation.infrastructure.storage.storage_adapter import StorageAdapter
from analogai.foundation.settings import config
from analogai.foundation.common.logging.app_logger import app_logger
import uuid
from typeguard import typechecked
import atexit
from analogai.foundation.common.design_patterns.singleton import Singleton
import time

class MongoDBStorageAdapterImpl(StorageAdapter, metaclass=Singleton):
    def __init__(self):
        app_logger.info("initializing mongo")
        self._validate_configuration()
        self.uri = config.get("MONGODB_URI")
        self.database_name = config.get("MONGODB_DATABASE")
        self.client: Optional[MongoClient] = None
        self.db = None
        self.max_retries = 3
        self.retry_delay = 2
        self._initialize_connection()
        atexit.register(self.close)
    
    def _validate_configuration(self):

        missing_configs = []
        
        if not config.get("MONGODB_URI"):
            missing_configs.append("MONGODB_URI")
        if not config.get("MONGODB_DATABASE"):
            missing_configs.append("MONGODB_DATABASE")
        
        if missing_configs:
            error_msg = f"MongoDB configuration is missing: {', '.join(missing_configs)}. Please check your environment variables."
            app_logger.error(error_msg)
            raise ValueError(error_msg)
        
        if not str(config.get("MONGODB_URI", "")).startswith(('mongodb://', 'mongodb+srv://')):
            error_msg = f"Invalid MongoDB URI format: {config.get('MONGODB_URI')}. Must start with mongodb:// or mongodb+srv://"
            app_logger.error(error_msg)
            raise ValueError(error_msg)
    
    def _initialize_connection(self):

        app_logger.debug(f"Attempting MongoDB connection with URI: {self.uri}")
        
        for attempt in range(self.max_retries):
            try:
                self._create_client()
                self._test_connection()
                app_logger.info(f"MongoDB connection established successfully at {self.uri}")
                return
                
            except ServerSelectionTimeoutError as e:
                error_msg = self._format_connection_error(e, attempt)
                if attempt < self.max_retries - 1:
                    app_logger.warning(f"{error_msg} Retrying in {self.retry_delay} seconds... (attempt {attempt + 1}/{self.max_retries})")
                    time.sleep(self.retry_delay)
                else:
                    app_logger.error(f"{error_msg} All retry attempts failed.")
                    raise ConnectionError(f"Failed to connect to MongoDB after {self.max_retries} attempts: {error_msg}")
                    
            except Exception as e:
                if "authentication" in str(e).lower() or "auth" in str(e).lower():
                    error_msg = f"MongoDB authentication failed. Check username and password. Error: {str(e)}"
                    app_logger.error(error_msg)
                    raise ConnectionError(error_msg)
                
            except ConfigurationError as e:
                error_msg = f"MongoDB configuration error: {str(e)}"
                app_logger.error(error_msg)
                raise ValueError(error_msg)
                
            except ConnectionFailure as e:
                error_msg = f"MongoDB connection failed: {str(e)}"
                app_logger.error(error_msg)
                if attempt < self.max_retries - 1:
                    app_logger.warning(f"Retrying in {self.retry_delay} seconds... (attempt {attempt + 1}/{self.max_retries})")
                    time.sleep(self.retry_delay)
                else:
                    raise ConnectionError(error_msg)
                    
            except Exception as e:
                error_msg = f"Unexpected error connecting to MongoDB: {str(e)}"
                app_logger.error(error_msg)
                if attempt < self.max_retries - 1:
                    app_logger.warning(f"Retrying in {self.retry_delay} seconds... (attempt {attempt + 1}/{self.max_retries})")
                    time.sleep(self.retry_delay)
                else:
                    raise ConnectionError(error_msg)
    
    def _create_client(self):

        try:
            self.client = MongoClient(
                self.uri,
                serverSelectionTimeoutMS=5000,
                connectTimeoutMS=5000,
                socketTimeoutMS=5000,
                maxPoolSize=50,
                retryWrites=True
            )
            self.db = self.client[self.database_name]
        except Exception as e:
            app_logger.error(f"Failed to create MongoDB client: {str(e)}")
            raise
    
    def _test_connection(self):

        try:
            self.client.admin.command('ping')
            app_logger.debug("MongoDB connection test successful")
        except Exception as e:
            app_logger.error(f"MongoDB connection test failed: {str(e)}")
            raise
    
    def _format_connection_error(self, error: Exception, attempt: int) -> str:

        error_str = str(error)
        
        if "ConnectionRefusedError" in error_str or "actively refused" in error_str:
            return f"MongoDB server is not running or not accessible at {self.uri}. Please ensure MongoDB is started and accessible."
        elif "timeout" in error_str.lower():
            return f"MongoDB connection timeout. The server can be overloaded or network issues exist. URI: {self.uri}"
        elif "authentication" in error_str.lower():
            return f"MongoDB authentication failed. Check credentials in URI: {self.uri}"
        else:
            return f"MongoDB connection failed: {error_str}"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self) -> None:

        if hasattr(self, 'client') and self.client:
            try:
                self.client.close()
                app_logger.info("MongoDB client connection closed successfully")
                self.client = None
                self.db = None
                if MongoDBStorageAdapterImpl in Singleton._instances:
                    Singleton._instances.pop(MongoDBStorageAdapterImpl, None)
            except Exception as e:
                app_logger.error(f"Error closing MongoDB client: {str(e)}")
                self.client = None
                self.db = None

    def is_connected(self) -> bool:

        if self.client is None:
            return False
        
        try:
            self.client.admin.command('ping')
            return True
        except Exception:
            return False
    
    def get_connection_info(self) -> dict:

        return {
            "uri": self.uri,
            "database": self.database_name,
            "is_connected": self.is_connected(),
            "client_active": self.client is not None
        }

    def __del__(self):

        try:
            self.close()
        except Exception:
            pass  # Ignore errors during cleanup

    def _normalize_document(self, doc: dict) -> dict:
        if not doc:
            return doc
        
        normalized = {}
        for key, value in doc.items():
            if key == '_id':
                continue
            
            if isinstance(value, int) and key.endswith('_id'):
                normalized[key] = str(value)
            elif isinstance(value, list):
                normalized[key] = [self._normalize_value(item) for item in value]
            elif isinstance(value, dict):
                normalized[key] = self._normalize_value(value)
            else:
                normalized[key] = value
        
        return normalized

    def _normalize_value(self, value):
        if isinstance(value, dict):
            normalized = {}
            for k, v in value.items():
                if isinstance(v, int) and k.endswith('_id'):
                    normalized[k] = str(v)
                elif isinstance(v, list):
                    normalized[k] = [self._normalize_value(item) for item in v]
                elif isinstance(v, dict):
                    normalized[k] = self._normalize_value(v)
                else:
                    normalized[k] = v
            return normalized
        elif isinstance(value, list):
            return [self._normalize_value(item) for item in value]
        else:
            return value

    @typechecked
    def store(self, collection: str, data: dict) -> dict:

        if not collection:
            raise ValueError("Collection name cannot be empty")
        if not data:
            raise ValueError("Data cannot be empty")
        
        try:
            self._ensure_connection()
            result = self.db[collection].replace_one({'id': data['id']}, data, upsert=True)
            app_logger.debug(f"Successfully stored document in mongodb collection {collection} (upserted: {result.upserted_id is not None})")
            stored = self.db[collection].find_one({'id': data['id']})
            return stored
        except OperationFailure as e:
            error_msg = f"Failed to store document in mongodb collection {collection}: {str(e)}"
            app_logger.error(error_msg)
            raise ConnectionError(error_msg)
        except Exception as e:
            error_msg = f"Unexpected error storing document in mongodb collection {collection}: {str(e)}"
            app_logger.error(error_msg)
            raise ConnectionError(error_msg)

    @typechecked
    def retrieve_by_id(self, collection: str, id: str) -> Optional[Any]:
        if not collection:
            raise ValueError("Collection name cannot be empty")
            
        app_logger.debug(f"Retrieving document with id: {id} from mongodb collection: {collection}")
        try:
            self._ensure_connection()
            result = self.db[collection].find_one({'id': id})
            app_logger.debug(f"Raw result from mongodb: {result}")
            if result:
                result = self._normalize_document(result)
            return result if result else None
        except OperationFailure as e:
            error_msg = f"Failed to retrieve document from mongodb collection {collection}: {str(e)}"
            app_logger.error(error_msg)
            raise ConnectionError(error_msg)
        except Exception as e:
            error_msg = f"Unexpected error retrieving document from mongodb collection {collection}: {str(e)}"
            app_logger.error(error_msg)
            raise ConnectionError(error_msg)

    @typechecked
    def retrieve_all(self, collection: str, query: Optional[dict] = None) -> List[Any]:

        if not collection:
            raise ValueError("Collection name cannot be empty")
            
        find_query = {}
        if query:
            find_query.update(query)
        try:
            self._ensure_connection()
            return [self._normalize_document(item) for item in self.db[collection].find(find_query)]
        except OperationFailure as e:
            error_msg = f"Failed to retrieve all documents from mongodb collection {collection}: {str(e)}"
            app_logger.error(error_msg)
            raise ConnectionError(error_msg)
        except Exception as e:
            error_msg = f"Unexpected error retrieving all documents from mongodb collection {collection}: {str(e)}"
            app_logger.error(error_msg)
            raise ConnectionError(error_msg)

    @typechecked
    def retrieve_by_attributes(self, collection: str, query: Dict[str, str]) -> List[Any]:

        if not collection:
            raise ValueError("Collection name cannot be empty")
        if not query:
            raise ValueError("Query cannot be empty")
            
        try:
            self._ensure_connection()
            return [self._normalize_document(item) for item in self.db[collection].find(query)]
        except OperationFailure as e:
            error_msg = f"Failed to retrieve documents by attributes from mongodb collection {collection}: {str(e)}"
            app_logger.error(error_msg)
            raise ConnectionError(error_msg)
        except Exception as e:
            error_msg = f"Unexpected error retrieving documents by attributes from mongodb collection {collection}: {str(e)}"
            app_logger.error(error_msg)
            raise ConnectionError(error_msg)

    @typechecked
    def delete(self, collection: str, id: str) -> bool:
        if not collection:
            raise ValueError("Collection name cannot be empty")
            
        try:
            self._ensure_connection()
            result = self.db[collection].delete_one({'id': id})
            if result.deleted_count == 0:
                app_logger.warning(f"Document with id {id} not found in collection {collection}")
                return False
            app_logger.debug(f"Successfully deleted document with id {id} from mongodb collection {collection}")
            return True
        except Exception as e:
            error_msg = f"Unexpected error deleting document from mongodb collection {collection}: {str(e)}"
            app_logger.error(error_msg)
            return False

    def clear(self) -> None:

        try:
            self._ensure_connection()
            collections = self.db.list_collection_names()
            if not collections:
                app_logger.info("No collections found to clear")
                return
                
            for collection in collections:
                result = self.db[collection].delete_many({})
                app_logger.debug(f"Cleared {result.deleted_count} documents from collection {collection}")
            app_logger.info(f"Successfully cleared all collections: {collections}")
        except OperationFailure as e:
            error_msg = f"Failed to clear mongodb collections: {str(e)}"
            app_logger.error(error_msg)
            raise ConnectionError(error_msg)
        except Exception as e:
            error_msg = f"Unexpected error clearing mongodb collections: {str(e)}"
            app_logger.error(error_msg)
            raise ConnectionError(error_msg)
    
    def _ensure_connection(self):

        if self.client is None or not self.is_connected():
            app_logger.warning("MongoDB connection lost, attempting to reconnect")
            self._initialize_connection()




from abc import ABC, abstractmethod
from typing import List, Optional

from analogai.foundation.modules.conditionals.hypothetical_statement import HypotheticalStatement


class HypotheticalStatementRepository(ABC):
    @abstractmethod
    def create(self, hypothetical_statement: HypotheticalStatement) -> HypotheticalStatement:
        pass

    @abstractmethod
    def find_by_id(self, hypothetical_statement_id: str) -> Optional[HypotheticalStatement]:
        pass

    @abstractmethod
    def find_by_agent_id(self, agent_id: str) -> List[HypotheticalStatement]:
        pass


from typing import Dict, List
from analogai.foundation.modules.agent.agent import Agent, Role
from analogai.foundation.modules.user.user import User
from analogai.foundation.common.enums.agent_openness import AgentOpenness


class AgentStorageMapper:
    @staticmethod
    def to_dict(agent: Agent) -> Dict:
        return {
            "id": agent.id,
            "name": agent.name,
            "knowledge_base": agent.knowledge_base,
            "creator_id": agent.creator.id,
            "openness": agent.openness.value,
            "roles": [
                {
                    "user_id": role.user.id,
                    "authority": role.authority
                } for role in agent.roles
            ]
        }

    @staticmethod
    def from_dict(data: Dict) -> Agent:
        creator_id = data.get("creator_id")
        if not creator_id:
            raise ValueError(f"Agent {data.get('id', 'unknown')} is missing creator_id")
        creator = User(id=creator_id, name="", email="")
        
        roles_doc = data.get("roles", [])
        roles: List[Role] = []
        if isinstance(roles_doc, list):
            for role_data in roles_doc:
                if role_data.get("user_id") and role_data.get("authority") is not None:
                    user = User(id=role_data["user_id"], name="", email="")
                    authority = float(role_data["authority"])
                    roles.append(Role(user=user, authority=authority))
        elif isinstance(roles_doc, dict) and roles_doc.get("user_id") and roles_doc.get("authority") is not None:
            user = User(id=roles_doc["user_id"], name="", email="")
            authority = float(roles_doc["authority"])
            roles.append(Role(user=user, authority=authority))
        
        openness_value = data.get("openness", AgentOpenness.OPEN.value)
        openness = AgentOpenness.OPEN
        try:
            openness = AgentOpenness(openness_value)
        except ValueError:
            openness = AgentOpenness.OPEN
        
        return Agent(
            id=data["id"],
            name=data.get("name", ""),
            knowledge_base=data.get("knowledge_base", ""),
            creator=creator,
            roles=roles,
            openness=openness
        )





from analogai.foundation.infrastructure.storage.storage_adapter import StorageAdapter
from typing import List, Dict, Any, Optional
from typeguard import typechecked
from analogai.foundation.common.logging.app_logger import app_logger
from analogai.foundation.common.design_patterns.singleton import Singleton
import atexit
import threading

class InMemoryStorageAdapterImpl(StorageAdapter, metaclass=Singleton):
    def __init__(self, db_path: Optional[str] = None):
        self._storage: Dict[str, Dict[str, Dict[str, Any]]] = {}
        self._lock = threading.Lock()
        atexit.register(self.close)

    def _get_collection(self, collection: str) -> Dict[str, Dict[str, Any]]:
        if collection not in self._storage:
            with self._lock:
                if collection not in self._storage:
                    self._storage[collection] = {}
        return self._storage[collection]

    def _match_nested(self, data: dict, path: str, value: Any) -> bool:
        parts = path.split('.')
        current = data
        for part in parts:
            if not isinstance(current, dict):
                return False
            current = current.get(part)
            if current is None:
                return False
        return current == value

    def _match_all(self, data: dict, query: Dict[str, Any]) -> bool:
        for attr_key, attr_value in query.items():
            if not self._match_nested(data, attr_key, attr_value):
                return False
        return True

    @typechecked
    def store(self, collection: str, data: Dict[str, Any]) -> Dict[str, Any]:
        collection_data = self._get_collection(collection)
        with self._lock:
            collection_data[data['id']] = data.copy()
        app_logger.debug(f"Stored object in {collection}: {data['id']}")
        return data

    @typechecked
    def retrieve_by_id(self, collection: str, id: str) -> Optional[Dict[str, Any]]:
        collection_data = self._get_collection(collection)
        with self._lock:
            result = collection_data.get(id)
        if result:
            app_logger.debug(f"Retrieved by ID {id} from {collection}: {result}")
            return result.copy()
        app_logger.debug(f"Retrieved by ID {id} from {collection}: None")
        return None

    @typechecked
    def retrieve_by_attributes(self, collection: str, query: Dict[str, Any]) -> List[Dict[str, Any]]:
        app_logger.debug(f"Retrieving by attributes from {collection}: {query}")
        collection_data = self._get_collection(collection)
        
        with self._lock:
            if not query:
                results = [doc.copy() for doc in collection_data.values()]
            else:
                results = [
                    doc.copy() for doc in collection_data.values()
                    if self._match_all(doc, query)
                ]
        
        app_logger.debug(f"Found {len(results)} results in {collection}")
        return results

    @typechecked
    def delete(self, collection: str, id: str) -> bool:
        collection_data = self._get_collection(collection)
        with self._lock:
            if id in collection_data:
                del collection_data[id]
                app_logger.debug(f"Deleted {id} from {collection}")
                return True
            app_logger.debug(f"Failed to delete {id} from {collection}")
            return False

    @typechecked
    def clear(self) -> bool:
        app_logger.debug("Clearing in-memory database")
        with self._lock:
            self._storage.clear()
        app_logger.debug("Cleared entire in-memory database")
        return True

    def close(self) -> None:
        self.clear()
        if InMemoryStorageAdapterImpl in Singleton._instances:
            Singleton._instances.pop(InMemoryStorageAdapterImpl, None)

    def __del__(self):
        try:
            self.close()
        except Exception:
            pass



from .in_memory_repository_strategy import InMemoryRepositoryStrategy
from .persistent_repository_strategy import PersistentRepositoryStrategy

__all__ = ['InMemoryRepositoryStrategy', 'PersistentRepositoryStrategy']




from abc import ABC, abstractmethod
from typing import List, Dict
from analogai.foundation.common.design_patterns.singleton import Singleton

class StorageAdapter(metaclass=Singleton):
    @abstractmethod
    def store(self, collection: str, data: dict) -> dict:
        pass

    @abstractmethod
    def retrieve_by_id(self, collection: str, id: str) -> dict | None:
        pass

    @abstractmethod
    def retrieve_by_attributes(self, collection: str, query: Dict[str, str]) -> List[dict]:
        pass

    @abstractmethod
    def delete(self, collection: str, id: str) -> bool:
        pass

    @abstractmethod
    def clear(self) -> bool:
        pass




import unittest

from analogai.foundation.common.design_patterns.singleton import Singleton

class Example(metaclass=Singleton):
    def __init__(self):
        self.value = 0

class TestSingleton(unittest.TestCase):
    def test_single_instance(self):
        a = Example()
        b = Example()
        self.assertIs(a, b)

    def test_state_shared(self):
        a = Example()
        b = Example()
        a.value = 42
        self.assertEqual(b.value, 42)

if __name__ == '__main__':
    unittest.main()





from abc import ABC, abstractmethod
from analogai.foundation.infrastructure.storage.storage_adapter import StorageAdapter
from analogai.foundation.infrastructure.storage.query_storage_adapter import QueryStorageAdapter

class StorageAdaptersFactory(ABC):
    @abstractmethod
    def get_storage_adapter(self) -> StorageAdapter:
        pass

    # @abstractmethod
    # def get_query_storage_adapter(self) -> QueryStorageAdapter:
    #     pass




from analogai.foundation.settings import config
from analogai.foundation.common.enums.environment import Environment
from analogai.foundation.infrastructure.storage.storage_adapters_factory import StorageAdaptersFactory
from analogai.foundation.infrastructure.factories.storage_adapters_factory_strategies.in_memory_storage_adapters_strategy import (
	InMemoryStorageAdaptersStrategy,
)
from analogai.foundation.infrastructure.factories.storage_adapters_factory_strategies.persistent_storage_adapters_strategy import (
	PersistentStorageAdaptersStrategy,
)


class StorageAdaptersFactoryImpl(StorageAdaptersFactory):
	def __init__(self, environment: str | None = None):
		env_value = environment or config.get("ENVIRONMENT")
		env = Environment(env_value)
		if env == Environment.PRODUCTION:
			self._strategy = PersistentStorageAdaptersStrategy()
		elif env in {Environment.DEVELOPMENT, Environment.TESTING}:
			self._strategy = InMemoryStorageAdaptersStrategy()
		else:
			raise ValueError(f"No storage adapter strategy defined for environment: {env}")

	def get_storage_adapter(self):
		return self._strategy.get_storage_adapter()


__all__ = [
	"StorageAdaptersFactoryImpl",
]



from abc import ABC, abstractmethod


class SimilarityService(ABC):
    @abstractmethod
    def similarity(self, text_a: str, text_b: str) -> float:
        raise NotImplementedError

    def is_similar(self, text_a: str, text_b: str, threshold: float) -> bool:
        return self.similarity(text_a, text_b) >= threshold



from analogai.foundation.settings import config
from analogai.foundation.extensions.contradictions.probability_variance_calculator import calculate_probability_variance


def is_belief_contradictory(belief) -> bool:
    probabilities = [(statement.validity, statement.probability) for statement in belief.statements]
    if not probabilities:
        raise ValueError("Cannot check contradiction for belief without statements")
    variance = calculate_probability_variance(probabilities)
    return variance > config.get("belief.variance", 0.1)



from .in_memory_storage_adapters_strategy import InMemoryStorageAdaptersStrategy
from .persistent_storage_adapters_strategy import PersistentStorageAdaptersStrategy

__all__ = ['InMemoryStorageAdaptersStrategy', 'PersistentStorageAdaptersStrategy']




import unittest
from analogai.foundation.modules.agent.agent import Agent, Role
from analogai.foundation.modules.user.user import User
from analogai.foundation.common.enums.agent_openness import AgentOpenness


class TestAgent(unittest.TestCase):
    def setUp(self):
        self.creator = User(id="creator_id", name="Creator", email="creator@example.com")
        self.user = User(id="user_id", name="Test User", email="user@example.com")
        self.role = Role(user=self.user, authority=0.8)

    def test_agent_default_initialization(self):
        agent = Agent(id="test_agent", creator=self.creator)

        self.assertEqual(agent.id, "test_agent")
        self.assertEqual(agent.knowledge_base, "")
        self.assertEqual(agent.creator, self.creator)
        self.assertEqual(len(agent.roles), 0)
        self.assertEqual(agent.openness, AgentOpenness.OPEN)

    def test_agent_full_initialization(self):
        agent = Agent(
            id="test_agent",
            knowledge_base="Test knowledge base",
            creator=self.creator,
            roles=[self.role]
        )

        self.assertEqual(agent.id, "test_agent")
        self.assertEqual(agent.knowledge_base, "Test knowledge base")
        self.assertEqual(agent.creator, self.creator)
        self.assertEqual(len(agent.roles), 1)
        self.assertEqual(agent.roles[0], self.role)

    def test_set_knowledge_base(self):
        agent = Agent(id="test_agent", creator=self.creator)

        agent.set_knowledge_base("New knowledge base")

        self.assertEqual(agent.knowledge_base, "New knowledge base")

    def test_agent_with_multiple_roles(self):
        user2 = User(id="user2_id", name="User Two", email="user2@example.com")
        role2 = Role(user=user2, authority=0.6)

        agent = Agent(
            id="test_agent",
            creator=self.creator,
            roles=[self.role, role2]
        )

        self.assertEqual(len(agent.roles), 2)
        self.assertEqual(agent.roles[0].user.id, "user_id")
        self.assertEqual(agent.roles[1].user.id, "user2_id")
        self.assertEqual(agent.roles[0].authority, 0.8)
        self.assertEqual(agent.roles[1].authority, 0.6)

    def test_agent_openness_functionality(self):
        agent = Agent(id="test_agent", creator=self.creator)

        self.assertTrue(agent.is_open())
        self.assertFalse(agent.is_closed())

        agent.set_openness(AgentOpenness.CLOSED)

        self.assertFalse(agent.is_open())
        self.assertTrue(agent.is_closed())

        agent.set_openness(AgentOpenness.OPEN)

        self.assertTrue(agent.is_open())
        self.assertFalse(agent.is_closed())

    def test_agent_full_initialization_with_all_fields(self):
        agent = Agent(
            id="test_agent",
            knowledge_base="Test knowledge base",
            creator=self.creator,
            roles=[self.role],
            openness=AgentOpenness.CLOSED
        )

        self.assertEqual(agent.id, "test_agent")
        self.assertEqual(agent.knowledge_base, "Test knowledge base")
        self.assertEqual(agent.creator, self.creator)
        self.assertEqual(len(agent.roles), 1)
        self.assertEqual(agent.openness, AgentOpenness.CLOSED)


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import Mock
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.common.utils.modifier_utils import ModifierMatcher
from analogai.foundation.modules.direct_statement.direct_statement import DirectStatement as Statement
from analogai.foundation.core.value_objects.time import Time
from analogai.foundation.core.value_objects.relation import Relation
from analogai.foundation.common.enums.deontic_modality import DeonticModality
from analogai.foundation.modules.notion.notion import Notion

class TestModifierMatcher(unittest.TestCase):
    
    def setUp(self):
        self.matcher = ModifierMatcher()
        self.subject_notion = Notion(user_id="user1", agent_id="agent1", caption="Source", id="source1")
        self.complement_notion = Notion(user_id="user1", agent_id="agent1", caption="Target", id="target1")
    
    def create_statement(self, location=None, from_time=None, to_time=None, deontic_modality=None):
        modifier = None
        if any([location, from_time, to_time, deontic_modality]):
            from analogai.foundation.common.utils.modifier_utils import ModifierBuilder
            builder = ModifierBuilder()
            if location:
                builder.with_location(location)
            if from_time:
                builder.with_from_time(
                    year=from_time.year,
                    month=from_time.month,
                    day=from_time.day,
                    hour=from_time.hour,
                    minute=from_time.minute
                )
            if to_time:
                builder.with_to_time(
                    year=to_time.year,
                    month=to_time.month,
                    day=to_time.day,
                    hour=to_time.hour,
                    minute=to_time.minute
                )
            if deontic_modality:
                builder.with_deontic_modality(deontic_modality)
            modifier = builder.build()
        return Statement(
            user_id="user1",
            agent_id="agent1",
            subject_notion=self.subject_notion,
            relation=Relation.from_string("is"),
            complement_notion=self.complement_notion,
            probability=0.8,
            validity=0.9,
            modifier=modifier
        )
    
    def test_matches_empty_context(self):
        statement = self.create_statement(location="New York")
        context = Modifier()
        
        result = self.matcher.matches(statement, context)
        
        self.assertTrue(result)
    
    def test_matches_location(self):
        statement = self.create_statement(location="New York")
        context = Modifier(location="New York")
        
        result = self.matcher.matches(statement, context)
        
        self.assertTrue(result)
    
    def test_does_not_match_location(self):
        statement = self.create_statement(location="New York")
        context = Modifier(location="London")
        
        result = self.matcher.matches(statement, context)
        
        self.assertFalse(result)
    
    def test_matches_deontic_modality(self):
        statement = self.create_statement(deontic_modality=DeonticModality.OBLIGATORY)
        context = Modifier(deontic_modality=DeonticModality.OBLIGATORY)
        
        result = self.matcher.matches(statement, context)
        
        self.assertTrue(result)
    
    def test_does_not_match_mismatched_deontic_modality(self):
        statement = self.create_statement(deontic_modality=DeonticModality.OBLIGATORY)
        context = Modifier(deontic_modality=DeonticModality.PERMITTED)
        
        result = self.matcher.matches(statement, context)
        
        self.assertFalse(result, "Statement with different deontic modality should not match")
    
    def test_does_not_match_when_search_has_deontic_but_statement_does_not(self):
        statement = self.create_statement()
        context = Modifier(deontic_modality=DeonticModality.OBLIGATORY)
        
        result = self.matcher.matches(statement, context)
        
        self.assertFalse(result, "Statement without deontic modality should not match search with deontic modality")
    
    def test_matches_when_search_has_no_deontic_but_statement_has_deontic(self):
        statement = self.create_statement(deontic_modality=DeonticModality.OBLIGATORY)
        context = Modifier()
        
        result = self.matcher.matches(statement, context)
        
        self.assertTrue(result, "Statement with deontic modality should match search without deontic modality")
    
    def test_matches_from_time_exact(self):
        from_time = Time(year=2024, month=1, day=1)
        statement = self.create_statement(from_time=from_time)
        context = Modifier(from_time=Time(year=2024, month=1, day=1))
        
        result = self.matcher.matches(statement, context)
        
        self.assertTrue(result)
    
    def test_matches_from_time_partial(self):
        from_time = Time(year=2024, month=1, day=1, hour=10)
        statement = self.create_statement(from_time=from_time)
        context = Modifier(from_time=Time(year=2024, month=1))
        
        result = self.matcher.matches(statement, context)
        
        self.assertTrue(result)
    
    def test_does_not_match_from_time(self):
        from_time = Time(year=2024, month=1, day=1)
        statement = self.create_statement(from_time=from_time)
        context = Modifier(from_time=Time(year=2024, month=2))
        
        result = self.matcher.matches(statement, context)
        
        self.assertFalse(result)
    
    def test_does_not_match_from_time_when_statement_has_none(self):
        statement = self.create_statement()
        context = Modifier(from_time=Time(year=2024))
        
        result = self.matcher.matches(statement, context)
        
        self.assertFalse(result)
    
    def test_matches_to_time_exact(self):
        to_time = Time(year=2024, month=12, day=31)
        statement = self.create_statement(to_time=to_time)
        context = Modifier(to_time=Time(year=2024, month=12, day=31))
        
        result = self.matcher.matches(statement, context)
        
        self.assertTrue(result)
    
    def test_matches_multiple_context(self):
        from_time = Time(year=2024, month=1)
        statement = self.create_statement(
            location="New York",
            from_time=from_time,
            deontic_modality=DeonticModality.OBLIGATORY
        )
        context = Modifier(
            location="New York",
            from_time=Time(year=2024, month=1),
            deontic_modality=DeonticModality.OBLIGATORY
        )
        
        result = self.matcher.matches(statement, context)
        
        self.assertTrue(result)
    
    def test_does_not_match_when_one_criterion_fails(self):
        from_time = Time(year=2024, month=1)
        statement = self.create_statement(
            location="New York",
            from_time=from_time,
            deontic_modality=DeonticModality.OBLIGATORY
        )
        context = Modifier(
            location="New York",
            from_time=Time(year=2024, month=2),
            deontic_modality=DeonticModality.OBLIGATORY
        )
        
        result = self.matcher.matches(statement, context)
        
        self.assertFalse(result)

if __name__ == '__main__':
    unittest.main()



from typing import List

from analogai.foundation.common.design_patterns.singleton import Singleton
from analogai.foundation.infrastructure.embeddings.base import EmbeddingProvider
from analogai.foundation.infrastructure.embeddings.provider_factory import create_embedding_provider

class EmbeddingService(metaclass=Singleton):
    def __init__(self):
        self._provider: EmbeddingProvider = create_embedding_provider()

    @property
    def dimension(self) -> int:
        dimension = self._provider.dimension
        if dimension is None:
            raise RuntimeError("Embedding dimension unavailable from provider.")
        return dimension

    def encode(self, text: str) -> List[float]:
        return self._provider.encode(text)

    def encode_batch(self, texts: List[str]) -> List[List[float]]:
        return self._provider.encode_batch(texts)


from dataclasses import dataclass
from typing import Optional, TYPE_CHECKING, List

from analogai.foundation.core.value_objects.time import Time
from analogai.foundation.core.value_objects.attribute import Attribute

from analogai.foundation.common.enums.deontic_modality import DeonticModality

@dataclass
class Modifier:
    location: Optional[str] = None
    deontic_modality: Optional[DeonticModality] = None
    from_time: Optional['Time'] = None
    to_time: Optional['Time'] = None
    attributes: Optional[List[Attribute]] = None
    quantity: Optional[float] = None



from datetime import datetime
from typing import Optional, Union, Any

from analogai.foundation.modules.shared.statement_base import StatementBase
from analogai.foundation.modules.shared.statement_validator import validate_probability_validity
from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects.relation import Relation
from analogai.foundation.common.utils.entity_formatters import (
    format_direct_statement_repr,
)


class DirectStatement(StatementBase):
    def __init__(
        self,
        user_id: str,
        agent_id: str,
        subject_notion: Notion,
        relation: Relation,
        complement_notion: Notion,
        id: Optional[str] = None,
        modifier: Optional[Modifier] = None,
        probability: Union[float, int] = 1.0,
        validity: Union[float, int] = 1.0,
        created_at: Optional[datetime] = None,
    ):
        if not user_id:
            raise ValueError("user_id must be a non-empty string")
        validate_probability_validity(probability, validity)
        super().__init__(
            agent_id=agent_id,
            subject_notion=subject_notion,
            complement_notion=complement_notion,
            relation=relation,
            id=id,
            modifier=modifier,
        )
        self.user_id = user_id
        self._probability = float(probability)
        self._validity = float(validity)
        self.created_at = created_at if created_at else datetime.now()

    @property
    def probability(self) -> float:
        return self._probability

    @probability.setter
    def probability(self, value: Union[float, int]) -> None:
        validate_probability_validity(value, self._validity)
        self._probability = float(value)

    @property
    def validity(self) -> float:
        return self._validity

    @validity.setter
    def validity(self, value: Union[float, int]) -> None:
        validate_probability_validity(self._probability, value)
        self._validity = float(value)

    def visit(self, visitor: Any) -> Any:
        return visitor.visit_statement(self)


    def __str__(self) -> str:
        return format_direct_statement_repr(self)

    def __repr__(self) -> str:
        return format_direct_statement_repr(self)


Statement = DirectStatement


from analogai.foundation.modules.shared.statement_base import StatementBase
from analogai.foundation.modules.shared.statement_validator import validate_probability_validity
from .direct_statement import DirectStatement, Statement
from .direct_statement_repository import StatementRepository
from .direct_statement_service import StatementService

__all__ = [
    "DirectStatement",
    "Statement",
    "StatementRepository",
    "StatementService",
    "StatementBase",
    "validate_probability_validity",
]


from .fallback_repository import FallbackRepository
from .in_memory_fallback_repository_impl import InMemoryFallbackRepositoryImpl
from .mongodb_fallback_repository_impl import MongoDBFallbackRepositoryImpl

__all__ = [
    "FallbackRepository",
    "InMemoryFallbackRepositoryImpl",
    "MongoDBFallbackRepositoryImpl",
]



from typing import Optional, List
import json
from analogai.foundation.modules.notion.notion_repository import NotionRepository
from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.foundation.infrastructure.storage.neo4j.neo4j_gateway import Neo4jGateway
from analogai.foundation.infrastructure.mappers import NotionStorageMapper
from analogai.foundation.settings import config
from analogai.foundation.common.logging.app_logger import app_logger


def _environment() -> str:
    value = config.get("ENVIRONMENT")
    return str(value).strip().lower()


def _use_vector_search() -> bool:
    if _environment() in {"test", "testing"}:
        return False
    return bool(config.get("vector.enabled", True))


def _similarity_cutoff() -> float:
    value = config.get("vector.cutoff", 0.4)
    try:
        return float(value)
    except (TypeError, ValueError):
        return 0.4

class Neo4jNotionRepositoryImpl(NotionRepository):
    VECTOR_INDEX_NAME = "notion_caption_embedding"
    DEFAULT_TOP_K = 10
    
    def __init__(self):
        self.neo4j_gateway = Neo4jGateway()
        self._embedding_service: Optional[object] = None
        if _use_vector_search():
            try:
                from analogai.foundation.infrastructure.services.embedding_service import EmbeddingService
                self._embedding_service = EmbeddingService()
            except Exception:
                self._embedding_service = None
        self._ensure_vector_index()
    
    def _get_author_id(self) -> Optional[str]:
        return None

    def _serialize_attributes(self, attributes: Optional[list]) -> Optional[str]:
        if attributes is None:
            return None
        try:
            return json.dumps(attributes)
        except (TypeError, ValueError):
            return None

    def _deserialize_attributes(self, raw: Optional[object]) -> Optional[list]:
        if raw is None:
            return None
        if isinstance(raw, list):
            return raw
        if isinstance(raw, str):
            try:
                return json.loads(raw)
            except (TypeError, ValueError):
                return None
        return None
    
    def _ensure_vector_index(self):
        if not _use_vector_search() or not self._embedding_service:
            return
        
        try:
            with self.neo4j_gateway.session() as session:
                result = session.run("SHOW INDEXES YIELD name WHERE name = $name", name=self.VECTOR_INDEX_NAME)
                if not result.single():
                    session.run(f"""
                        CREATE VECTOR INDEX {self.VECTOR_INDEX_NAME} IF NOT EXISTS
                        FOR (n:Notion)
                        ON n.embedding
                        OPTIONS {{
                            indexConfig: {{
                                `vector.dimensions`: {self._embedding_service.dimension},
                                `vector.similarity_function`: 'cosine'
                            }}
                        }}
                    """)
        except Exception as e:
            pass
    
    def create(self, notion: Notion) -> Optional[Notion]:
        try:
            data = NotionStorageMapper.to_dict(notion)
            attributes_json = self._serialize_attributes(data.get("attributes"))
            
            if _use_vector_search() and self._embedding_service:
                embedding = self._embedding_service.encode(notion.caption)
                query = """
                    MERGE (n:Notion {
                        caption: $caption,
                        user_id: $user_id,
                        agent_id: $agent_id
                    })
                    ON CREATE SET 
                        n.id = $id,
                        n.embedding = $embedding,
                        n.attributes_json = $attributes_json
                    ON MATCH SET 
                        n.embedding = $embedding,
                        n.attributes_json = $attributes_json
                    RETURN n.id as notion_id
                """
                params = {
                    "id": data["id"], 
                    "caption": data["caption"], 
                    "user_id": data["user_id"], 
                    "agent_id": data["agent_id"],
                    "embedding": embedding,
                    "attributes_json": attributes_json,
                }
            else:
                query = """
                    MERGE (n:Notion {
                        caption: $caption,
                        user_id: $user_id,
                        agent_id: $agent_id
                    })
                    ON CREATE SET 
                        n.id = $id,
                        n.attributes_json = $attributes_json
                    ON MATCH SET
                        n.attributes_json = $attributes_json
                    RETURN n.id as notion_id
                """
                params = {
                    "id": data["id"], 
                    "caption": data["caption"], 
                    "user_id": data["user_id"], 
                    "agent_id": data["agent_id"],
                    "attributes_json": attributes_json,
                }
            
            with self.neo4j_gateway.session() as session:
                result = session.run(query, **params)
                record = result.single()
                if record:
                    actual_id = record["notion_id"]
                    if actual_id != data["id"]:
                        notion.id = actual_id
                result.consume()
            return notion
        except Exception as e:
            app_logger.exception("Failed to create notion in Neo4j", exc_info=e)
            return None
    
    def delete(self, notion: Notion) -> None:
        author_id = self._get_author_id()
        if not author_id:
            return
        try:
            with self.neo4j_gateway.session() as session:
                session.run(
                    "MATCH (n:Notion {id: $id}) "
                    "DETACH DELETE n",
                    id=notion.id
                )
        except Exception as e:
            pass
    
    def find(self, agent_id: str, expression: NotionExpression, context_ids: Optional[List[str]] = None, similarity_threshold: float = _similarity_cutoff()) -> Optional[Notion]:
        if not _use_vector_search() or not self._embedding_service:
            return self._find_exact_match(agent_id, expression)
        
        query_embedding = self._embedding_service.encode(expression.caption)
        
        query = f"""
            CALL db.index.vector.queryNodes('{self.VECTOR_INDEX_NAME}', $top_k, $query_embedding)
            YIELD node, score
            WHERE node.agent_id = $agent_id AND score >= $threshold
            RETURN node, score
            ORDER BY score DESC
            LIMIT 1
        """
        
        params = {
            "query_embedding": query_embedding,
            "agent_id": agent_id,
            "top_k": self.DEFAULT_TOP_K,
            "threshold": similarity_threshold
        }
        
        try:
            with self.neo4j_gateway.session() as session:
                result = session.run(query, **params)
                record = result.single()
                if record:
                    notion_data = dict(record["node"])
                    attributes = self._deserialize_attributes(notion_data.get("attributes_json") or notion_data.get("attributes"))
                    return NotionStorageMapper.from_dict({
                        "id": notion_data.get("id"),
                        "user_id": notion_data.get("user_id", ""),
                        "agent_id": notion_data.get("agent_id", ""),
                        "caption": notion_data.get("caption", ""),
                        "attributes": attributes,
                    })
                return None
        except Exception as e:
            return self._find_exact_match(agent_id, expression)
    
    def _find_exact_match(self, agent_id: str, expression: NotionExpression) -> Optional[Notion]:
        query = (
            "MATCH (n:Notion) "
            "WHERE toLower(n.caption) = toLower($caption) "
            "AND n.agent_id = $agent_id "
            "RETURN n LIMIT 1"
        )
        params = {"caption": expression.caption, "agent_id": agent_id}
        
        try:
            with self.neo4j_gateway.session() as session:
                result = session.run(query, **params)
                for record in result:
                    notion_data = dict(record["n"])
                    attributes = self._deserialize_attributes(notion_data.get("attributes_json") or notion_data.get("attributes"))
                    return NotionStorageMapper.from_dict({
                        "id": notion_data.get("id"),
                        "user_id": notion_data.get("user_id", ""),
                        "agent_id": notion_data.get("agent_id", ""),
                        "caption": notion_data.get("caption", ""),
                        "attributes": attributes,
                    })
                return None
        except Exception as e:
            return None
    
    def find_all_by_agent(self, agent_id: str) -> List[Notion]:
        try:
            with self.neo4j_gateway.session() as session:
                result = session.run(
                    "MATCH (n:Notion) WHERE n.agent_id = $agent_id RETURN n",
                    agent_id=agent_id
                )
                notions = []
                for record in result:
                    notion_data = dict(record["n"])
                    attributes = self._deserialize_attributes(notion_data.get("attributes_json") or notion_data.get("attributes"))
                    notions.append(NotionStorageMapper.from_dict({
                        "id": notion_data.get("id"),
                        "user_id": notion_data.get("user_id", ""),
                        "agent_id": notion_data.get("agent_id", ""),
                        "caption": notion_data.get("caption", ""),
                        "attributes": attributes,
                    }))
                return notions
        except Exception as e:
            return []

    def find_by_id(self, notion_id: str) -> Optional[Notion]:
        author_id = self._get_author_id()
        try:
            with self.neo4j_gateway.session() as session:
                result = session.run(
                    "MATCH (n:Notion {id: $id}) "
                    "RETURN n",
                    id=notion_id
                )
                record = result.single()
                if record:
                    notion_data = dict(record["n"])
                    attributes = self._deserialize_attributes(notion_data.get("attributes_json") or notion_data.get("attributes"))
                    return NotionStorageMapper.from_dict({
                        "id": notion_data.get("id"),
                        "user_id": notion_data.get("user_id", ""),
                        "agent_id": notion_data.get("agent_id", ""),
                        "caption": notion_data.get("caption", ""),
                        "attributes": attributes,
                    })
                return None
        except Exception as e:
            return None
    
    def exists(self, agent_id: str, expression: NotionExpression) -> bool:
        return self.find(agent_id, expression, None) is not None
    
    def search(self, expression: NotionExpression, similarity_threshold: float = _similarity_cutoff(), limit: int = 10) -> List[Notion]:
        if not _use_vector_search() or not self._embedding_service:
            return self._search_exact_match(expression)
        
        query_embedding = self._embedding_service.encode(expression.caption)
        
        query = f"""
            CALL db.index.vector.queryNodes('{self.VECTOR_INDEX_NAME}', $top_k, $query_embedding)
            YIELD node, score
            WHERE score >= $threshold
            RETURN node, score
            ORDER BY score DESC
            LIMIT $limit
        """
        
        params = {
            "query_embedding": query_embedding,
            "top_k": max(self.DEFAULT_TOP_K, limit),
            "threshold": similarity_threshold,
            "limit": limit
        }
        
        try:
            with self.neo4j_gateway.session() as session:
                result = session.run(query, **params)
                notions = []
                for record in result:
                    notion_data = dict(record["node"])
                    attributes = self._deserialize_attributes(notion_data.get("attributes_json") or notion_data.get("attributes"))
                    notions.append(NotionStorageMapper.from_dict({
                        "id": notion_data.get("id"),
                        "user_id": notion_data.get("user_id", ""),
                        "agent_id": notion_data.get("agent_id", ""),
                        "caption": notion_data.get("caption", ""),
                        "attributes": attributes,
                    }))
                return notions
        except Exception as e:
            return self._search_exact_match(expression)
    
    def _search_exact_match(self, expression: NotionExpression) -> List[Notion]:
        query = (
            "MATCH (n:Notion) "
            "WHERE toLower(n.caption) = toLower($caption) "
            "RETURN n"
        )
        params = {"caption": expression.caption}
        
        try:
            with self.neo4j_gateway.session() as session:
                result = session.run(query, **params)
                notions = []
                for record in result:
                    notion_data = dict(record["n"])
                    attributes = self._deserialize_attributes(notion_data.get("attributes_json") or notion_data.get("attributes"))
                    notions.append(NotionStorageMapper.from_dict({
                        "id": notion_data.get("id"),
                        "user_id": notion_data.get("user_id", ""),
                        "agent_id": notion_data.get("agent_id", ""),
                        "caption": notion_data.get("caption", ""),
                        "attributes": attributes,
                    }))
                return notions
        except Exception as e:
            return []


from unittest import TestCase
import unittest
from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects.relation import Relation
from analogai.foundation.common.utils.modifier_utils import ModifierBuilder
from analogai.foundation.core.value_objects.time import Time
from analogai.foundation.common.enums.deontic_modality import DeonticModality
from analogai.foundation.tests.functional.services.setup import ServicesTestConfig

class TestBeliefServiceModifierFunctional(TestCase):
    def setUp(self):
        self.config = ServicesTestConfig()
        self.user = self.config.user
        self.agent = self.config.agent
        self.belief_service = self.config.belief_service
        self.statement_service = self.config.statement_service
        self.subject_notion = Notion(user_id=self.user.id, agent_id=self.agent.id, caption="human", id="human_id")
        self.complement_notion = Notion(user_id=self.user.id, agent_id=self.agent.id, caption="mortal", id="mortal_id")
        self.config.notion_repository.create(self.subject_notion)
        self.config.notion_repository.create(self.complement_notion)

    def tearDown(self):
        self.config.tear_down()

    def test_belief_preserves_modifier_through_storage(self):
        modifier = Modifier(
            location="Berlin",
            from_time=Time(year=2024, month=6, day=15),
            to_time=Time(year=2024, month=12, day=31),
            deontic_modality=DeonticModality.PERMITTED
        )
        self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            complement_notion=self.complement_notion,
            relation=Relation.from_string("is"),
            probability=0.8,
            validity=0.9,
            modifier=modifier
        )
        belief = self.belief_service.find(
            self.subject_notion, self.complement_notion, Relation.from_string("is"),
            modifier=modifier
        )
        self.assertIsNotNone(belief)
        self.assertEqual(belief.location, "Berlin")
        self.assertIsNotNone(belief.from_time)
        self.assertEqual(belief.from_time.year, 2024)
        self.assertEqual(belief.from_time.month, 6)
        self.assertEqual(belief.from_time.day, 15)
        self.assertIsNotNone(belief.to_time)
        self.assertEqual(belief.to_time.year, 2024)
        self.assertEqual(belief.to_time.month, 12)
        self.assertEqual(belief.deontic_modality, DeonticModality.PERMITTED)
        retrieved_belief = self.belief_service.find(
            self.subject_notion, self.complement_notion, Relation.from_string("is"),
            modifier=modifier
        )
        self.assertIsNotNone(retrieved_belief)
        if retrieved_belief is None:
            self.fail("Expected belief to be found")
        self.assertEqual(retrieved_belief.location, "Berlin")
        self.assertEqual(retrieved_belief.deontic_modality, DeonticModality.PERMITTED)

    def test_search_beliefs_by_modifier_location(self):
        statement1 = self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            complement_notion=self.complement_notion,
            relation=Relation.from_string("is"),
            probability=0.8,
            validity=0.9,
            modifier=Modifier(location="Vienna")
        )
        other_notion = Notion(user_id=self.user.id, agent_id=self.agent.id, caption="animal", id="animal_id")
        self.config.notion_repository.create(other_notion)
        statement2 = self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            complement_notion=other_notion,
            relation=Relation.from_string("is"),
            probability=0.7,
            validity=0.85,
            modifier=Modifier(location="Prague")
        )
        results = self.belief_service.search_by_subject_notion(
            self.subject_notion,
            modifier=Modifier(location="Vienna")
        )
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].location, "Vienna")
        self.assertEqual(len(results[0].statements), 1)
        self.assertEqual(results[0].statements[0].id, statement1.id)

    def test_search_beliefs_by_modifier_deontic_modality(self):
        self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            complement_notion=self.complement_notion,
            relation=Relation.from_string("is"),
            probability=0.8,
            validity=0.9,
            modifier=Modifier(deontic_modality=DeonticModality.OBLIGATORY)
        )
        other_notion = Notion(user_id=self.user.id, agent_id=self.agent.id, caption="animal", id="animal_id")
        self.config.notion_repository.create(other_notion)
        self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            complement_notion=other_notion,
            relation=Relation.from_string("is"),
            probability=0.7,
            validity=0.85,
            modifier=Modifier(deontic_modality=DeonticModality.PERMITTED)
        )
        results = self.belief_service.search_by_subject_notion(
            self.subject_notion,
            modifier=Modifier(deontic_modality=DeonticModality.OBLIGATORY)
        )
        self.assertEqual(len(results), 1, "Deontic modality filtering should return matching beliefs")
        result_ids = {r.id for r in results}
        statement1_belief = self.belief_service.find(
            self.subject_notion, self.complement_notion, Relation.from_string("is"),
            modifier=Modifier(deontic_modality=DeonticModality.OBLIGATORY)
        )
        statement2_belief = self.belief_service.find(
            self.subject_notion, other_notion, Relation.from_string("is"),
            modifier=Modifier(deontic_modality=DeonticModality.PERMITTED)
        )
        self.assertIn(statement1_belief.id, result_ids)
        self.assertNotIn(statement2_belief.id, result_ids)

    def test_search_beliefs_by_modifier_time(self):
        from_time = Time(year=2024, month=3, day=1)
        statement1 = self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            complement_notion=self.complement_notion,
            relation=Relation.from_string("is"),
            probability=0.8,
            validity=0.9,
            modifier=Modifier(from_time=from_time)
        )
        other_notion = Notion(user_id=self.user.id, agent_id=self.agent.id, caption="animal", id="animal_id")
        self.config.notion_repository.create(other_notion)
        statement2 = self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            complement_notion=other_notion,
            relation=Relation.from_string("is"),
            probability=0.7,
            validity=0.85,
            modifier=Modifier(from_time=Time(year=2024, month=6))
        )
        results = self.belief_service.search_by_subject_notion(
            self.subject_notion,
            modifier=Modifier(from_time=Time(year=2024, month=3))
        )
        self.assertEqual(len(results), 1)
        self.assertIsNotNone(results[0].from_time)
        self.assertEqual(results[0].from_time.year, 2024)
        self.assertEqual(results[0].from_time.month, 3)
        self.assertEqual(len(results[0].statements), 1)
        self.assertEqual(results[0].statements[0].id, statement1.id)

    def test_search_beliefs_by_modifier_location_and_time(self):
        statement1 = self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            complement_notion=self.complement_notion,
            relation=Relation.from_string("is"),
            probability=0.8,
            validity=0.9,
            modifier=Modifier(location="Lisbon", from_time=Time(year=2024, month=4))
        )
        other_notion = Notion(user_id=self.user.id, agent_id=self.agent.id, caption="animal", id="animal_id")
        self.config.notion_repository.create(other_notion)
        self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            complement_notion=other_notion,
            relation=Relation.from_string("is"),
            probability=0.7,
            validity=0.85,
            modifier=Modifier(location="Lisbon", from_time=Time(year=2024, month=5))
        )
        self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            complement_notion=other_notion,
            relation=Relation.from_string("is"),
            probability=0.7,
            validity=0.85,
            modifier=Modifier(location="Porto", from_time=Time(year=2024, month=4))
        )

        results = self.belief_service.search_by_subject_notion(
            self.subject_notion,
            modifier=Modifier(location="Lisbon", from_time=Time(year=2024, month=4))
        )

        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].location, "Lisbon")
        self.assertIsNotNone(results[0].from_time)
        self.assertEqual(results[0].from_time.month, 4)
        self.assertEqual(len(results[0].statements), 1)
        self.assertEqual(results[0].statements[0].id, statement1.id)

    def test_search_beliefs_by_modifier_multiple_properties(self):
        self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            complement_notion=self.complement_notion,
            relation=Relation.from_string("is"),
            probability=0.8,
            validity=0.9,
            modifier=ModifierBuilder()
                .with_location("Madrid")
                .with_deontic_modality(DeonticModality.FORBIDDEN)
                .with_from_time(year=2024, month=5)
                .build()
        )
        other_notion = Notion(user_id=self.user.id, agent_id=self.agent.id, caption="animal", id="animal_id")
        self.config.notion_repository.create(other_notion)
        self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            complement_notion=other_notion,
            relation=Relation.from_string("is"),
            probability=0.7,
            validity=0.85,
            modifier=ModifierBuilder()
                .with_location("Barcelona")
                .with_deontic_modality(DeonticModality.FORBIDDEN)
                .with_from_time(year=2024, month=5)
                .build()
        )
        results = self.belief_service.search_by_subject_notion(
            self.subject_notion,
            modifier=ModifierBuilder()
                .with_location("Madrid")
                .with_deontic_modality(DeonticModality.FORBIDDEN)
                .with_from_time(year=2024, month=5)
                .build()
        )
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].location, "Madrid")
        self.assertEqual(results[0].deontic_modality, DeonticModality.FORBIDDEN)
        self.assertIsNotNone(results[0].from_time)
        self.assertEqual(results[0].from_time.year, 2024)
        self.assertEqual(results[0].from_time.month, 5)

    def test_belief_modifier_persists_after_adding_multiple_statements(self):
        modifier = Modifier(
            location="Rome",
            deontic_modality=DeonticModality.OBLIGATORY
        )
        self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            complement_notion=self.complement_notion,
            relation=Relation.from_string("is"),
            probability=0.8,
            validity=0.9,
            modifier=modifier
        )
        from analogai.foundation.modules.user.user import User
        other_user = User(id="other_user_id", name="Other User", email="other@example.com")
        self.statement_service.create(
            user_id=other_user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            complement_notion=self.complement_notion,
            relation=Relation.from_string("is"),
            probability=0.7,
            validity=0.85,
            modifier=modifier
        )
        belief = self.belief_service.find(
            self.subject_notion, self.complement_notion, Relation.from_string("is"),
            modifier=modifier
        )
        self.assertIsNotNone(belief)
        self.assertEqual(len(belief.statements), 2)
        self.assertEqual(belief.location, "Rome")
        self.assertEqual(belief.deontic_modality, DeonticModality.OBLIGATORY)
        retrieved_belief = self.belief_service.find(
            self.subject_notion, self.complement_notion, Relation.from_string("is"),
            modifier=modifier
        )
        self.assertIsNotNone(retrieved_belief)
        if retrieved_belief is None:
            self.fail("Expected belief to be found")
        self.assertEqual(retrieved_belief.location, "Rome")
        self.assertEqual(retrieved_belief.deontic_modality, DeonticModality.OBLIGATORY)

    def test_search_beliefs_by_modifier_returns_empty_when_no_match(self):
        self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            complement_notion=self.complement_notion,
            relation=Relation.from_string("is"),
            probability=0.8,
            validity=0.9,
            modifier=Modifier(location="Amsterdam")
        )
        results = self.belief_service.search_by_subject_notion(
            self.subject_notion,
            modifier=Modifier(location="Brussels")
        )
        self.assertEqual(len(results), 0)

    def test_find_with_modifier_returns_matching_belief(self):
        modal_a = Modifier(location="Dublin")
        self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            complement_notion=self.complement_notion,
            relation=Relation.from_string("is"),
            probability=0.8,
            validity=0.9,
            modifier=modal_a
        )
        belief_by_modal = self.belief_service.find(
            self.subject_notion, self.complement_notion, Relation.from_string("is"), modifier=modal_a
        )
        self.assertIsNotNone(belief_by_modal)
        self.assertEqual(belief_by_modal.location, "Dublin")
        self.assertEqual(len(belief_by_modal.statements), 1)

    def test_find_with_modifier_returns_none_when_no_match(self):
        self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            complement_notion=self.complement_notion,
            relation=Relation.from_string("is"),
            probability=0.8,
            validity=0.9,
            modifier=Modifier(location="Cork")
        )
        other_modal = Modifier(location="Galway")
        belief = self.belief_service.find(
            self.subject_notion, self.complement_notion, Relation.from_string("is"), modifier=other_modal
        )
        self.assertIsNone(belief)

if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod
from typing import List, Optional, Any
from datetime import datetime

from analogai.foundation.modules.belief.belief import Belief
from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects.relation import Relation


class BeliefRepository(ABC):

    @abstractmethod
    def save(self, belief: Belief) -> Belief:
        pass

    @abstractmethod
    def find(self, subject_notion: Notion, complement_notion: Notion,
             relation: Relation,
             modifier: Optional[Modifier] = None,
             quantity: Optional[float] = None) -> Optional[Belief]:
        pass

    @abstractmethod
    def delete(self, belief: Belief) -> None:
        pass

    @abstractmethod
    def search_by_notions(self, subject_notion: Notion, complement_notion: Notion,
                          modifier: Optional[Modifier] = None) -> List[Belief]:
        pass

    @abstractmethod
    def search_by_subject_notion(self, subject_notion: Notion,
                                modifier: Optional[Modifier] = None) -> List[Belief]:
        pass

    @abstractmethod
    def search_by_complement_notion(self, complement_notion: Notion,
                                modifier: Optional[Modifier] = None) -> List[Belief]:
        pass

    @abstractmethod
    def find_by_agent_id(self, agent_id: str) -> List[Belief]:
        pass

    @abstractmethod
    def get_recently_updated(self, agent_id: str, since: datetime) -> List[Belief]:
        pass

    @abstractmethod
    def clear_cache(self, belief_id: Optional[str] = None) -> None:
        pass

    @abstractmethod
    def search_by_modifier(self, modifier: Modifier) -> List[Belief]:
        pass

    @abstractmethod
    def find_belief_by_statement_id(self, statement_id: str) -> Optional[Belief]:
        pass

    def set_belief_service(self, belief_service: Any) -> None:
        pass

    def set_time_service(self, time_service: Any) -> None:
        pass


import unittest
from unittest.mock import Mock, patch
from analogai.foundation.modules.agent.agent import Agent, Role
from analogai.foundation.modules.user.user import User
from analogai.foundation.common.enums.agent_openness import AgentOpenness
from analogai.foundation.infrastructure.repositories.mongodb.mongodb_agent_repository_impl import MongoDBAgentRepositoryImpl
from analogai.foundation.infrastructure.storage.storage_adapter import StorageAdapter
from analogai.foundation.infrastructure.tests.repositories.mongodb.test_config import MongoDBEnvironmentCheckMixin

class TestMongoDBAgentRepositoryImpl(MongoDBEnvironmentCheckMixin, unittest.TestCase):
    
    def setUp(self):
        self.mock_storage_adapter = Mock(spec=StorageAdapter)
        self.repository = MongoDBAgentRepositoryImpl(self.mock_storage_adapter)
        
        self.test_user = User(id="user_123", name="Test User", email="test@example.com")
        
        self.test_agent = Agent(
            id="agent_456",
            knowledge_base="Test knowledge base",
            creator=self.test_user,
            roles=[]
        )
    
    def test_find_by_id_existing_agent(self):
        doc = {
            "id": "agent_456",
            "knowledge_base": "Test knowledge base",
            "creator_id": "user_123",
            "openness": "open",
            "roles": []
        }
        self.mock_storage_adapter.retrieve_by_id.return_value = doc
        
        result = self.repository.find_by_id("agent_456")
        
        self.assertIsNotNone(result)
        self.assertEqual(result.id, "agent_456")
        self.assertEqual(result.knowledge_base, "Test knowledge base")
        self.assertEqual(result.creator.id, "user_123")
        self.mock_storage_adapter.retrieve_by_id.assert_called_once_with(
            MongoDBAgentRepositoryImpl.COLLECTION_NAME, 
            "agent_456"
        )
    
    def test_find_by_id_nonexistent_agent(self):
        self.mock_storage_adapter.retrieve_by_id.return_value = None
        
        result = self.repository.find_by_id("nonexistent")
        
        self.assertIsNone(result)
        self.mock_storage_adapter.retrieve_by_id.assert_called_once_with(
            MongoDBAgentRepositoryImpl.COLLECTION_NAME, 
            "nonexistent"
        )
    
    def test_create_agent(self):
        self.repository.create(self.test_agent)
        
        self.mock_storage_adapter.store.assert_called_once()
        call_args = self.mock_storage_adapter.store.call_args
        self.assertEqual(call_args[0][0], MongoDBAgentRepositoryImpl.COLLECTION_NAME)
        
        stored_data = call_args[0][1]
        self.assertEqual(stored_data["id"], "agent_456")
        self.assertEqual(stored_data["knowledge_base"], "Test knowledge base")
        self.assertEqual(stored_data["creator_id"], "user_123")
    
    def test_update_agent(self):
        self.repository.update(self.test_agent)
        
        self.mock_storage_adapter.store.assert_called_once()
        call_args = self.mock_storage_adapter.store.call_args
        self.assertEqual(call_args[0][0], MongoDBAgentRepositoryImpl.COLLECTION_NAME)
    
    def test_delete_agent(self):
        self.mock_storage_adapter.delete.return_value = True
        
        result = self.repository.delete("agent_456")
        
        self.assertTrue(result)
        self.mock_storage_adapter.delete.assert_called_once_with(
            MongoDBAgentRepositoryImpl.COLLECTION_NAME, 
            "agent_456"
        )
    
    def test_find_by_creator_id_with_agents(self):
        doc1 = {
            "id": "agent_456",
            "knowledge_base": "Test knowledge base",
            "creator_id": "user_123",
            "openness": "open",
            "roles": []
        }
        
        self.mock_storage_adapter.retrieve_by_attributes.return_value = [doc1]
        
        result = self.repository.find_by_creator_id("user_123")
        
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].id, "agent_456")
        self.assertEqual(result[0].creator.id, "user_123")
        self.mock_storage_adapter.retrieve_by_attributes.assert_called_once_with(
            MongoDBAgentRepositoryImpl.COLLECTION_NAME,
            {"creator_id": "user_123"}
        )
    
    def test_find_by_creator_id_no_agents(self):
        self.mock_storage_adapter.retrieve_by_attributes.return_value = []
        
        result = self.repository.find_by_creator_id("user_123")
        
        self.assertEqual(len(result), 0)
        self.mock_storage_adapter.retrieve_by_attributes.assert_called_once_with(
            MongoDBAgentRepositoryImpl.COLLECTION_NAME,
            {"creator_id": "user_123"}
        )
    

if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod
from typing import Any, Awaitable, Callable
from analogai.foundation.modules.user.user import User
from analogai.foundation.extensions.authentication.models import TokenPayload

AuthenticationDependency = Callable[[Any], Awaitable[User]]
TokenExtractionDependency = Callable[[Any], Awaitable[TokenPayload]]
ApiTokenAuthenticationDependency = Callable[[Any], Awaitable[User]]

class AuthenticationFactoryInterface(ABC):
    @abstractmethod
    def create_authentication_dependency(self) -> AuthenticationDependency:
        pass

    @abstractmethod
    def create_token_extraction_dependency(self) -> TokenExtractionDependency:
        pass

    @abstractmethod
    def create_api_token_authentication_dependency(self) -> ApiTokenAuthenticationDependency:
        pass




from analogai.foundation.extensions.fallback.services.fallback_service import FallbackService
from analogai.foundation.extensions.fallback.repositories import InMemoryFallbackRepositoryImpl, MongoDBFallbackRepositoryImpl
from analogai.foundation.common.enums.environment import Environment
from analogai.foundation.settings import config

class FallbackServiceFactory:
    def create(self) -> FallbackService:
        env = Environment(config.get("ENVIRONMENT"))
        if env == Environment.PRODUCTION:
            repository = MongoDBFallbackRepositoryImpl()
        else:
            repository = InMemoryFallbackRepositoryImpl()
        return FallbackService(repository)



from typing import Optional, Dict, Any
from analogai.foundation.settings import config
from analogai.foundation.common.enums.criterion import Criterion
from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.modules.belief.belief import Belief
from analogai.foundation.modules.categorical_deduction.categorical_deduction import CategoricalDeduction
from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.foundation.common.utils.expression_assembler import ExpressionAssembler


class ExpressionFactory:
    def __init__(self):
        self._assembler = ExpressionAssembler(self._maybe_lemmatize_text, self._lemmatize_criteria)

    def _maybe_lemmatize_text(self, text: str) -> str:
        if not config.get("lemmatize", False):
            return text
        from analogai.foundation.common.utils.lemmatization import lemmatize_text
        return lemmatize_text(text)

    def _lemmatize_criteria(self, criteria: Dict[Criterion, Any]) -> Dict[Criterion, Any]:
        if not criteria:
            return {}
        normalized = dict(criteria)
        subject = normalized.get(Criterion.SubjectCaption)
        if subject is not None:
            normalized[Criterion.SubjectCaption] = self._maybe_lemmatize_text(str(subject))
        object_caption = normalized.get(Criterion.ObjectCaption)
        if object_caption is not None:
            normalized[Criterion.ObjectCaption] = self._maybe_lemmatize_text(str(object_caption))
        return normalized
    
    def create_notion_expression(self, caption: str, attributes=None) -> NotionExpression:
        return self._assembler.build_notion_expression(caption, attributes)
    
    def propagate_notion_expression(self, notion: Notion) -> NotionExpression:
        return self._assembler.build_notion_expression_from_notion(notion)
    
    def propagate_belief_expression(self, belief: Belief) -> "BeliefExpression":
        return self._assembler.build_belief_expression(belief)
    
    def propagate_categorical_deduction_expression(self, inference: CategoricalDeduction) -> "CategoricalDeductionExpression":
        return self._assembler.build_categorical_deduction_expression(inference)

    def propagate_hypothetical_statement_expression(self, hypothetical: "HypotheticalStatement") -> "HypotheticalStatementExpression":
        return self._assembler.build_hypothetical_statement_expression(hypothetical)




from abc import ABC, abstractmethod
from typing import Awaitable, Callable
from fastapi import Request
from analogai.foundation.extensions.authorization.models import AuthorizationContext
from analogai.foundation.extensions.authorization.core.enums.operation import AuthorizationOperation

AuthorizationDependency = Callable[..., Awaitable[AuthorizationContext]]

class AuthorizationFactoryInterface(ABC):
    @abstractmethod
    def create_authorization_dependency(
        self,
        operation: AuthorizationOperation = AuthorizationOperation.CHAT,
    ) -> AuthorizationDependency:
        pass

    @abstractmethod
    def create_api_authorization_dependency(
        self,
        operation: AuthorizationOperation = AuthorizationOperation.CHAT,
    ) -> AuthorizationDependency:
        pass




from __future__ import annotations

import json
from typing import Any, Dict
from urllib import request, error


def post_json(url: str, payload: Dict[str, Any], headers: Dict[str, str], timeout: int = 30) -> Dict[str, Any]:
    data = json.dumps(payload).encode("utf-8")
    req = request.Request(url, data=data, headers=headers, method="POST")
    try:
        with request.urlopen(req, timeout=timeout) as response:
            body = response.read().decode("utf-8")
            return json.loads(body) if body else {}
    except error.HTTPError as exc:
        detail = exc.read().decode("utf-8") if exc.fp else ""
        raise RuntimeError(f"Embedding API request failed with status {exc.code}: {detail}") from exc
    except error.URLError as exc:
        raise RuntimeError(f"Embedding API request failed: {exc.reason}") from exc



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional
from fastapi import HTTPException, Request, status
from analogai.foundation.modules.user.user import User
from analogai.foundation.modules.agent.agent import Agent
from analogai.foundation.modules.agent.agent_service import AgentService
from analogai.foundation.extensions.authentication.interfaces.authentication_factory_interface import (
	AuthenticationFactoryInterface,
)
from analogai.foundation.extensions.authorization.interfaces.authorization_factory_interface import (
	AuthorizationDependency,
	AuthorizationFactoryInterface,
)
from analogai.foundation.extensions.authorization.core import AuthorizationOperation, AuthorizationService
from analogai.foundation.extensions.authorization.models import AuthorizationContext
from analogai.foundation.common.logging.app_logger import app_logger
from analogai.foundation.extensions.authorization.authority_policy import (
	AuthorityPolicy,
	DefaultAuthorityPolicy,
)
from analogai.foundation.extensions.authorization.error_mappers import AuthorizationHttpErrorMapper
from analogai.foundation.extensions.authentication.authentication_factory import AuthenticationFactory
from analogai.foundation.setup.factories.repository_selector import create_repository_strategy


@dataclass(frozen=True)
class ApiAuthorizationInput:
	request: Request
	user: User
	agent_id: str
	operation_override: AuthorizationOperation
	authorization_service: AuthorizationService
	agent_service: AgentService
	authority_policy: AuthorityPolicy
	error_mapper: AuthorizationHttpErrorMapper


class ApiAuthorizationContextResolver:
	auth_type: str

	def resolve(self, data: ApiAuthorizationInput) -> AuthorizationContext:
		raise NotImplementedError

	def _resolve_agent_id(self, data: ApiAuthorizationInput) -> str:
		token_agent_id = getattr(data.request.state, "api_agent_id", None)
		if token_agent_id and data.agent_id and token_agent_id != data.agent_id:
			raise HTTPException(
				status_code=status.HTTP_403_FORBIDDEN,
				detail="Agent mismatch for API token",
			)
		resolved_agent_id = token_agent_id or data.agent_id
		if not resolved_agent_id:
			raise HTTPException(
				status_code=status.HTTP_400_BAD_REQUEST,
				detail="Missing agent_id",
			)
		return resolved_agent_id


class UserApiAuthorizationResolver(ApiAuthorizationContextResolver):
	auth_type = "user"

	def resolve(self, data: ApiAuthorizationInput) -> AuthorizationContext:
		resolved_agent_id = self._resolve_agent_id(data)
		try:
			agent = data.authorization_service.find_and_authorize_agent(
				data.user,
				resolved_agent_id,
				data.operation_override,
			)
		except ValueError as error:
			raise data.error_mapper.map(error) from error

		authority = data.authority_policy.resolve_api(self.auth_type)
		app_logger.info(
			"API authorization context resolved: user_id=%s agent_id=%s user_email=%s authority=%s auth_type=%s",
			data.user.id if data.user else None,
			agent.id,
			data.user.email if data.user else None,
			authority,
			self.auth_type,
		)
		return AuthorizationContext(
			user=data.user,
			agent=agent,
			authority=authority,
			operation=data.operation_override,
		)


class SystemApiAuthorizationResolver(ApiAuthorizationContextResolver):
	auth_type = "system"

	def resolve(self, data: ApiAuthorizationInput) -> AuthorizationContext:
		resolved_agent_id = self._resolve_agent_id(data)
		agent = data.agent_service.find(resolved_agent_id)
		if not agent:
			raise HTTPException(
				status_code=status.HTTP_404_NOT_FOUND,
				detail="Agent not found",
			)

		authority = data.authority_policy.resolve_api(self.auth_type)
		app_logger.info(
			"API authorization context resolved: user_id=%s agent_id=%s user_email=%s authority=%s auth_type=%s",
			data.user.id if data.user else None,
			agent.id,
			data.user.email if data.user else None,
			authority,
			self.auth_type,
		)
		return AuthorizationContext(
			user=data.user,
			agent=agent,
			authority=authority,
			operation=data.operation_override,
		)


class GuestApiAuthorizationResolver(SystemApiAuthorizationResolver):
	auth_type = "guest"


class AuthorizationFactory(AuthorizationFactoryInterface):
	def __init__(self):
		self._authorization_service: Optional[AuthorizationService] = None
		self._authentication_factory: Optional[AuthenticationFactoryInterface] = None
		self._agent_service: Optional[AgentService] = None
		self._authority_policy: Optional[AuthorityPolicy] = None
		self._error_mapper: Optional[AuthorizationHttpErrorMapper] = None
		self._api_resolvers: Optional[dict[str, ApiAuthorizationContextResolver]] = None

	def create_authorization_dependency(
		self,
		operation: AuthorizationOperation = AuthorizationOperation.CHAT,
	) -> AuthorizationDependency:
		authentication_dependency = self._get_authentication_factory().create_authentication_dependency()
		authorization_service = self._get_authorization_service()
		agent_service = self._get_agent_service()

		async def dependency(
			request: Request,
			agent_id: str,
			operation_override: AuthorizationOperation = operation,
		) -> AuthorizationContext:
			user = await authentication_dependency(request)

			if getattr(request.state, "api_auth_type", None) in {"user", "system", "guest"}:
				return self._resolve_api_authorization_context(
					request=request,
					user=user,
					agent_id=agent_id,
					operation_override=operation_override,
					authorization_service=authorization_service,
					agent_service=agent_service,
				)

			try:
				agent = authorization_service.find_and_authorize_agent(
					user,
					agent_id,
					operation_override,
				)
			except ValueError as error:
				raise self._get_error_mapper().map(error) from error

			authority = self._extract_authority(user, agent)
			app_logger.info(
				"Authorization context resolved: user_id=%s agent_id=%s user_email=%s authority=%s",
				user.id if user else None,
				agent.id,
				user.email if user else None,
				authority,
			)
			return AuthorizationContext(
				user=user,
				agent=agent,
				authority=authority,
				operation=operation_override,
			)

		return dependency

	def create_api_authorization_dependency(
		self,
		operation: AuthorizationOperation = AuthorizationOperation.CHAT,
	) -> AuthorizationDependency:
		authentication_dependency = (
			self._get_authentication_factory().create_api_token_authentication_dependency()
		)
		authorization_service = self._get_authorization_service()
		agent_service = self._get_agent_service()

		async def dependency(
			request: Request,
			agent_id: str,
			operation_override: AuthorizationOperation = operation,
		) -> AuthorizationContext:
			user = await authentication_dependency(request)
			return self._resolve_api_authorization_context(
				request=request,
				user=user,
				agent_id=agent_id,
				operation_override=operation_override,
				authorization_service=authorization_service,
				agent_service=agent_service,
			)

		return dependency

	def _get_authorization_service(self) -> AuthorizationService:
		if not self._authorization_service:
			self._authorization_service = AuthorizationService(self._get_agent_service())
		return self._authorization_service

	def _get_agent_service(self) -> AgentService:
		if not self._agent_service:
			from analogai.foundation.extensions.authorization import authorization_factory as authz_factory

			repository = authz_factory.create_repository_strategy().get_agent_repository()
			self._agent_service = AgentService(repository)
		return self._agent_service

	def _get_authentication_factory(self) -> AuthenticationFactoryInterface:
		if not self._authentication_factory:
			from analogai.foundation.extensions.authorization import authorization_factory as authz_factory

			self._authentication_factory = authz_factory.AuthenticationFactory()
		return self._authentication_factory

	def _get_authority_policy(self) -> AuthorityPolicy:
		if not self._authority_policy:
			self._authority_policy = DefaultAuthorityPolicy()
		return self._authority_policy

	def _get_error_mapper(self) -> AuthorizationHttpErrorMapper:
		if not self._error_mapper:
			self._error_mapper = AuthorizationHttpErrorMapper.default()
		return self._error_mapper

	def _get_api_resolvers(self) -> dict[str, ApiAuthorizationContextResolver]:
		if not self._api_resolvers:
			resolvers = (
				UserApiAuthorizationResolver(),
				SystemApiAuthorizationResolver(),
				GuestApiAuthorizationResolver(),
			)
			self._api_resolvers = {resolver.auth_type: resolver for resolver in resolvers}
		return self._api_resolvers

	def _extract_authority(self, user: User, agent: Agent) -> float:
		return self._get_authority_policy().resolve_role(user, agent)

	def _resolve_api_authorization_context(
		self,
		request: Request,
		user: User,
		agent_id: str,
		operation_override: AuthorizationOperation,
		authorization_service: AuthorizationService,
		agent_service: AgentService,
	) -> AuthorizationContext:
		auth_type = getattr(request.state, "api_auth_type", "system")
		resolver = self._get_api_resolvers().get(auth_type)
		if not resolver:
			raise HTTPException(
				status_code=status.HTTP_401_UNAUTHORIZED,
				detail="Invalid auth_type; expected 'user', 'system', or 'guest'",
			)

		data = ApiAuthorizationInput(
			request=request,
			user=user,
			agent_id=agent_id,
			operation_override=operation_override,
			authorization_service=authorization_service,
			agent_service=agent_service,
			authority_policy=self._get_authority_policy(),
			error_mapper=self._get_error_mapper(),
		)
		return resolver.resolve(data)


__all__ = ["AuthorizationFactory", "AuthenticationFactory", "create_repository_strategy"]




import unittest
from analogai.foundation.modules.belief.utils.belief_certainties_calculator_util import calculate_belief_certainties


class TestBeliefCertaintiesCalculatorUtil(unittest.TestCase):
    def test_calculate_belief_certainties_single_statement(self):
        certainties = [(1.0, 0.8)]
        probability, validity = calculate_belief_certainties(certainties)

        self.assertEqual(probability, 0.8)
        self.assertEqual(validity, 1.0)

    def test_calculate_belief_certainties_multiple_statements(self):
        certainties = [(1.0, 0.8), (0.5, 0.6)]

        probability, validity = calculate_belief_certainties(certainties)

        self.assertAlmostEqual(probability, 0.73, places=2)
        self.assertAlmostEqual(validity, 0.83, places=2)

    def test_calculate_belief_certainties_empty_list(self):
        certainties = []

        with self.assertRaises(ValueError):
            calculate_belief_certainties(certainties)

    def test_calculate_belief_certainties_zero_validity(self):
        certainties = [(0.0, 0.8)]

        with self.assertRaises(ValueError):
            calculate_belief_certainties(certainties)

    def test_calculate_belief_certainties_mixed_validities(self):
        certainties = [(1.0, 0.8), (0.2, 0.6), (0.5, 0.4)]

        probability, validity = calculate_belief_certainties(certainties)

        self.assertAlmostEqual(probability, 0.66, places=2)
        self.assertAlmostEqual(validity, 0.76, places=2)


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod
from .observable_service import ObservableService


class NotionObserver(ABC):
    @abstractmethod
    def on_notion_created(self, notion):
        pass


class NotionObservableService(ObservableService):
    def notify_notion_created(self, notion) -> None:
        for observer in self._observers:
            try:
                observer.on_notion_created(notion)
            except (AttributeError, TypeError):
                pass


import unittest
from unittest.mock import Mock
from analogai.foundation.modules.agent.agent import Agent, Role
from analogai.foundation.modules.user.user import User
from analogai.foundation.infrastructure.repositories.in_memory.in_memory_agent_repository_impl import InMemoryAgentRepositoryImpl
from analogai.foundation.infrastructure.storage.storage_adapter import StorageAdapter
from analogai.foundation.infrastructure.mappers import AgentStorageMapper

class TestInMemoryAgentRepositoryImpl(unittest.TestCase):
    
    def setUp(self):
        self.mock_storage_adapter = Mock(spec=StorageAdapter)
        self.repository = InMemoryAgentRepositoryImpl(storage_adapter=self.mock_storage_adapter)
        
        self.test_user = User(id="user_123", name="Test User", email="test@example.com")
        self.test_agent = Agent(
            id="agent_456",
            knowledge_base="Test knowledge base",
            creator=self.test_user,
            roles=[]
        )
    
    def test_create_agent(self):
        self.repository.create(self.test_agent)
        
        expected_data = AgentStorageMapper.to_dict(self.test_agent)
        self.mock_storage_adapter.store.assert_called_once_with(
            InMemoryAgentRepositoryImpl.COLLECTION_NAME, 
            expected_data
        )
    
    def test_find_by_id_existing_agent(self):
        agent_data = AgentStorageMapper.to_dict(self.test_agent)
        self.mock_storage_adapter.retrieve_by_id.return_value = agent_data
        
        result = self.repository.find_by_id("agent_456")
        
        self.assertEqual(result.id, self.test_agent.id)
        self.assertEqual(result.knowledge_base, self.test_agent.knowledge_base)
        self.mock_storage_adapter.retrieve_by_id.assert_called_once_with(
            InMemoryAgentRepositoryImpl.COLLECTION_NAME, 
            "agent_456"
        )
    
    def test_find_by_id_nonexistent_agent(self):
        self.mock_storage_adapter.retrieve_by_id.return_value = None
        
        result = self.repository.find_by_id("nonexistent")
        
        self.assertIsNone(result)
        self.mock_storage_adapter.retrieve_by_id.assert_called_once_with(
            InMemoryAgentRepositoryImpl.COLLECTION_NAME, 
            "nonexistent"
        )
    
    def test_update_agent(self):
        self.repository.update(self.test_agent)
        
        expected_data = AgentStorageMapper.to_dict(self.test_agent)
        self.mock_storage_adapter.store.assert_called_once_with(
            InMemoryAgentRepositoryImpl.COLLECTION_NAME, 
            expected_data
        )
    
    def test_delete_agent_success(self):
        self.mock_storage_adapter.delete.return_value = True
        
        result = self.repository.delete("agent_456")
        
        self.assertTrue(result)
        self.mock_storage_adapter.delete.assert_called_once_with(
            InMemoryAgentRepositoryImpl.COLLECTION_NAME, 
            "agent_456"
        )
    
    def test_delete_agent_failure(self):
        self.mock_storage_adapter.delete.return_value = False
        
        result = self.repository.delete("nonexistent")
        
        self.assertFalse(result)
        self.mock_storage_adapter.delete.assert_called_once_with(
            InMemoryAgentRepositoryImpl.COLLECTION_NAME, 
            "nonexistent"
        )
    
    def test_find_by_creator_id_with_agents(self):
        agent_data = AgentStorageMapper.to_dict(self.test_agent)
        self.mock_storage_adapter.retrieve_by_attributes.return_value = [agent_data]
        
        result = self.repository.find_by_creator_id("user_123")
        
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].id, self.test_agent.id)
        self.mock_storage_adapter.retrieve_by_attributes.assert_called_once_with(
            InMemoryAgentRepositoryImpl.COLLECTION_NAME,
            {"creator_id": "user_123"}
        )
    
    def test_find_by_creator_id_no_agents(self):
        self.mock_storage_adapter.retrieve_by_attributes.return_value = []
        
        result = self.repository.find_by_creator_id("user_123")
        
        self.assertEqual(len(result), 0)
        self.mock_storage_adapter.retrieve_by_attributes.assert_called_once_with(
            InMemoryAgentRepositoryImpl.COLLECTION_NAME,
            {"creator_id": "user_123"}
        )
    
    def test_find_by_creator_id_multiple_agents(self):
        another_agent = Agent(
            id="agent_789",
            knowledge_base="Another knowledge base",
            creator=self.test_user,
            roles=[]
        )
        
        agent_data1 = AgentStorageMapper.to_dict(self.test_agent)
        agent_data2 = AgentStorageMapper.to_dict(another_agent)
        self.mock_storage_adapter.retrieve_by_attributes.return_value = [agent_data1, agent_data2]
        
        result = self.repository.find_by_creator_id("user_123")
        
        self.assertEqual(len(result), 2)
        ids = {r.id for r in result}
        self.assertIn(self.test_agent.id, ids)
        self.assertIn(another_agent.id, ids)
        self.mock_storage_adapter.retrieve_by_attributes.assert_called_once_with(
            InMemoryAgentRepositoryImpl.COLLECTION_NAME,
            {"creator_id": "user_123"}
        )
    
    def test_find_by_creator_id_agent_without_creator(self):
        agent_data = AgentStorageMapper.to_dict(self.test_agent)
        self.mock_storage_adapter.retrieve_by_attributes.return_value = [agent_data]
        
        result = self.repository.find_by_creator_id("user_123")
        
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].id, self.test_agent.id)
        self.mock_storage_adapter.retrieve_by_attributes.assert_called_once_with(
            InMemoryAgentRepositoryImpl.COLLECTION_NAME,
            {"creator_id": "user_123"}
        )

if __name__ == '__main__':
    unittest.main()



from analogai.foundation.modules.conditionals.conditional_deduction_service import ConditionalDeductionService
from .belief_service_factory import BeliefServiceFactory
from .hypothetical_statement_service_factory import HypotheticalStatementServiceFactory
from .notion_service_factory import NotionServiceFactory


class ConditionalDeductionServiceFactory:
    def __init__(
        self,
        belief_service=None,
        notion_service=None,
        hypothetical_statement_service=None,
        environment: str | None = None,
    ):
        self.belief_service = belief_service or BeliefServiceFactory(environment=environment).get_service()
        self.notion_service = notion_service or NotionServiceFactory(environment=environment).get_service()
        self.hypothetical_statement_service = (
            hypothetical_statement_service or HypotheticalStatementServiceFactory(environment=environment).get_service()
        )
        self.service = ConditionalDeductionService(
            hypothetical_statement_service=self.hypothetical_statement_service,
            notion_service=self.notion_service,
            belief_service=self.belief_service,
        )

    def get_service(self) -> ConditionalDeductionService:
        return self.service


from typing import Union


def validate_probability_validity(probability: Union[float, int], validity: Union[float, int]) -> None:
    prob = float(probability) if probability is not None else 1.0
    valid = float(validity) if validity is not None else 1.0
    if not 0 <= prob <= 1:
        raise ValueError("probability must be between 0 and 1")
    if not 0 <= valid <= 1:
        raise ValueError("validity must be between 0 and 1")


from abc import ABC, abstractmethod
from typing import Optional, List
from analogai.foundation.modules.agent.agent import Agent

class AgentRepository(ABC):
    @abstractmethod
    def find_by_id(self, id: str) -> Optional[Agent]:
        pass

    @abstractmethod
    def create(self, agent: Agent) -> Agent:
        pass

    @abstractmethod
    def update(self, agent: Agent) -> Agent:
        pass

    @abstractmethod
    def delete(self, id: str) -> bool:
        pass

    @abstractmethod
    def find_by_creator_id(self, creator_id: str) -> List[Agent]:
        pass


from typing import Optional, Any

from analogai.foundation.modules.shared.statement_base import StatementBase
from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects.relation import Relation
from analogai.foundation.common.logging.app_logger import app_logger
from analogai.foundation.modules.belief.belief import Belief
from analogai.foundation.common.utils.entity_formatters import format_categorical_deduction_repr


class CategoricalDeduction(StatementBase):
    def __init__(
        self,
        agent_id: str,
        major_premise: Belief,
        minor_premise: Belief,
        subject_notion: Notion,
        complement_notion: Notion,
        relation: Relation,
        id: Optional[str] = None,
        modifier: Optional[Modifier] = None,
    ):
        super().__init__(
            agent_id=agent_id,
            subject_notion=subject_notion,
            complement_notion=complement_notion,
            relation=relation,
            id=id,
            modifier=modifier,
        )
        self._major_premise = major_premise
        self._minor_premise = minor_premise
        self._log_creation()

    @property
    def major_premise(self):
        return self._major_premise

    @property
    def minor_premise(self):
        return self._minor_premise

    @property
    def probability(self) -> float:
        return round(self.major_premise.probability * self.minor_premise.probability, 2)

    @property
    def validity(self) -> float:
        return min(self.major_premise.validity, self.minor_premise.validity)

    def visit(self, visitor: Any) -> Any:
        return visitor.visit_categorical_deduction(self)

    def _log_creation(self):
        from analogai.foundation.common.utils.statement_serializers import serialize_statement
        try:
            probability = float(self.probability)
            serialized = serialize_statement(
                subject_notion_caption=self.subject_notion.caption,
                complement_notion_caption=self.complement_notion.caption,
                relation=self.relation,
                probability=probability,
                modifier=self.modifier,
                subject_attributes=self.subject_notion.attributes,
                complement_attributes=self.complement_notion.attributes,
            )
            app_logger.debug(f"Created CategoricalDeduction: {serialized} (id: {self.id})")
        except (ValueError, AttributeError, TypeError):
            app_logger.debug(f"Created CategoricalDeduction (id: {self.id})")

    def __repr__(self) -> str:
        return format_categorical_deduction_repr(self)


class StateMachine:
    def __init__(self, context):
        self.context = context
        self.current_state = None
        self.previous_state = None

    def state_to(self, next_state):
        self.previous_state = self.current_state
        self.current_state = next_state
        return self.current_state.execute()

    def set_context(self, context):
        self.context = context




class State:
	def __init__(self, state_machine):
		self.state_machine = state_machine

	@property
	def context(self):
		return self.state_machine.context
		
	@context.setter
	def context(self, context):
		self.state_machine.context = context

	def set_context(self, context):
		self.state_machine.context = context



import unittest

from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects.time import Time
from analogai.foundation.modules.categorical_deduction.categorical_deduction import CategoricalDeduction
from analogai.foundation.core.value_objects.relation import Relation
from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.tests.functional.services.setup import ServicesTestConfig


class TestCategoricalDeductionServiceFunctional(unittest.TestCase):
    def setUp(self):
        self.config = ServicesTestConfig()
        self.user = self.config.user
        self.agent = self.config.agent
        self.belief_service = self.config.belief_service
        self.statement_service = self.config.statement_service
        self.categorical_deduction_service = self.config.categorical_deduction_service
        self.notion_repository = self.config.notion_repository

    def tearDown(self):
        self.config.tear_down()

    def _create_notion(self, caption: str, notion_id: str) -> Notion:
        notion = Notion(user_id=self.user.id, agent_id=self.agent.id, caption=caption, id=notion_id)
        self.notion_repository.create(notion)
        return notion

    def _create_statement(self, subject: Notion, complement_notion: Notion, relation: Relation, modifier: Modifier | None = None):
        return self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=subject,
            complement_notion=complement_notion,
            relation=relation,
            probability=1.0,
            validity=1.0,
            modifier=modifier,
        )

    def _deductions_from_belief(self, belief):
        return [stmt for stmt in belief.statements if isinstance(stmt, CategoricalDeduction)]

    def test_live_categorical_deduction_builds_belief(self):
        socrates = self._create_notion("socrates", "socrates_id")
        human = self._create_notion("human", "human_id")
        mortal = self._create_notion("mortal", "mortal_id")

        self._create_statement(socrates, human, Relation.from_string("is"))
        self._create_statement(human, mortal, Relation.from_string("is"))

        beliefs_by_key = self.belief_service._base_beliefs_for_agent(self.agent.id, None, None)
        self.categorical_deduction_service.resolve_deductions(
            beliefs_by_key,
            modifier=None,
            max_depth=5,
        )
        belief_key = self.belief_service._belief_key_from_values(
            agent_id=self.agent.id,
            subject_id=socrates.id,
            object_id=mortal.id,
            relation=Relation.from_string("is"),
            modifier=None,
        )
        belief = beliefs_by_key.get(belief_key)
        self.assertIsNotNone(belief)
        deductions = self._deductions_from_belief(belief)
        self.assertEqual(len(deductions), 1)

        major_key = self.belief_service._belief_key_from_values(
            agent_id=self.agent.id,
            subject_id=human.id,
            object_id=mortal.id,
            relation=Relation.from_string("is"),
            modifier=None,
        )
        minor_key = self.belief_service._belief_key_from_values(
            agent_id=self.agent.id,
            subject_id=socrates.id,
            object_id=human.id,
            relation=Relation.from_string("is"),
            modifier=None,
        )
        major = beliefs_by_key.get(major_key)
        minor = beliefs_by_key.get(minor_key)
        self.assertIsNotNone(major)
        self.assertIsNotNone(minor)
        if major is None or minor is None:
            self.fail("Expected major and minor premises to be available")
        self.assertEqual(deductions[0].major_premise, major)
        self.assertEqual(deductions[0].minor_premise, minor)

    def test_find_deductions_by_agent_id(self):
        socrates = self._create_notion("socrates", "socrates_id")
        human = self._create_notion("human", "human_id")
        mortal = self._create_notion("mortal", "mortal_id")

        self._create_statement(socrates, human, Relation.from_string("is"))
        self._create_statement(human, mortal, Relation.from_string("is"))

        deductions = self.categorical_deduction_service.find_deductions_by_agent_id(
            agent_id=self.agent.id,
            max_depth=5,
            limit=10,
            offset=0,
        )
        target_relation = Relation.from_string("is")
        matching = [
            deduction
            for deduction in deductions
            if isinstance(deduction, CategoricalDeduction)
            and deduction.subject_notion.id == socrates.id
            and deduction.complement_notion.id == mortal.id
            and deduction.relation == target_relation
        ]
        self.assertEqual(len(matching), 1)
        deduction = matching[0]
        self.assertIsNotNone(deduction.major_premise)
        self.assertIsNotNone(deduction.minor_premise)
        self.assertEqual(deduction.major_premise.subject_notion.id, human.id)
        self.assertEqual(deduction.major_premise.complement_notion.id, mortal.id)
        self.assertEqual(deduction.minor_premise.subject_notion.id, socrates.id)
        self.assertEqual(deduction.minor_premise.complement_notion.id, human.id)

    def test_multiple_deductions_with_different_premises(self):
        socrates = self._create_notion("socrates", "socrates_id")
        human = self._create_notion("human", "human_id")
        man = self._create_notion("man", "man_id")
        mortal = self._create_notion("mortal", "mortal_id")

        self._create_statement(socrates, human, Relation.from_string("is"))
        self._create_statement(human, mortal, Relation.from_string("is"))
        self._create_statement(socrates, man, Relation.from_string("is"))
        self._create_statement(man, mortal, Relation.from_string("is"))

        beliefs_by_key = self.belief_service._base_beliefs_for_agent(self.agent.id, None, None)
        self.categorical_deduction_service.resolve_deductions(
            beliefs_by_key,
            modifier=None,
            max_depth=5,
        )
        belief_key = self.belief_service._belief_key_from_values(
            agent_id=self.agent.id,
            subject_id=socrates.id,
            object_id=mortal.id,
            relation=Relation.from_string("is"),
            modifier=None,
        )
        belief = beliefs_by_key.get(belief_key)
        self.assertIsNotNone(belief)
        deductions = self._deductions_from_belief(belief)
        self.assertEqual(len(deductions), 2)

        major_human = beliefs_by_key.get(self.belief_service._belief_key_from_values(
            agent_id=self.agent.id,
            subject_id=human.id,
            object_id=mortal.id,
            relation=Relation.from_string("is"),
            modifier=None,
        ))
        minor_human = beliefs_by_key.get(self.belief_service._belief_key_from_values(
            agent_id=self.agent.id,
            subject_id=socrates.id,
            object_id=human.id,
            relation=Relation.from_string("is"),
            modifier=None,
        ))
        major_man = beliefs_by_key.get(self.belief_service._belief_key_from_values(
            agent_id=self.agent.id,
            subject_id=man.id,
            object_id=mortal.id,
            relation=Relation.from_string("is"),
            modifier=None,
        ))
        minor_man = beliefs_by_key.get(self.belief_service._belief_key_from_values(
            agent_id=self.agent.id,
            subject_id=socrates.id,
            object_id=man.id,
            relation=Relation.from_string("is"),
            modifier=None,
        ))
        self.assertIsNotNone(major_human)
        self.assertIsNotNone(minor_human)
        self.assertIsNotNone(major_man)
        self.assertIsNotNone(minor_man)
        if major_human is None or minor_human is None or major_man is None or minor_man is None:
            self.fail("Expected major/minor premises to be available")

        pairs = {(d.major_premise, d.minor_premise) for d in deductions}
        self.assertIn((major_human, minor_human), pairs)
        self.assertIn((major_man, minor_man), pairs)

    def test_multi_step_categorical_deduction_depth(self):
        socrates = self._create_notion("socrates", "socrates_id")
        philosopher = self._create_notion("philosopher", "philosopher_id")
        human = self._create_notion("human", "human_id")
        mortal = self._create_notion("mortal", "mortal_id")

        self._create_statement(socrates, philosopher, Relation.from_string("is"))
        self._create_statement(philosopher, human, Relation.from_string("is"))
        self._create_statement(human, mortal, Relation.from_string("is"))

        beliefs_by_key = self.belief_service._base_beliefs_for_agent(self.agent.id, None, None)
        self.categorical_deduction_service.resolve_deductions(
            beliefs_by_key,
            modifier=None,
            max_depth=5,
        )
        belief_key = self.belief_service._belief_key_from_values(
            agent_id=self.agent.id,
            subject_id=socrates.id,
            object_id=mortal.id,
            relation=Relation.from_string("is"),
            modifier=None,
        )
        belief = beliefs_by_key.get(belief_key)
        self.assertIsNotNone(belief)
        deductions = self._deductions_from_belief(belief)
        self.assertGreaterEqual(len(deductions), 1)

    def test_modifier_scopes_categorical_deductions(self):
        socrates = self._create_notion("socrates", "socrates_id")
        human = self._create_notion("human", "human_id")
        mortal = self._create_notion("mortal", "mortal_id")

        athens = Modifier(location="Athens")
        sparta = Modifier(location="Sparta")

        self._create_statement(socrates, human, Relation.from_string("is"), modifier=athens)
        self._create_statement(human, mortal, Relation.from_string("is"), modifier=athens)
        self._create_statement(socrates, human, Relation.from_string("is"), modifier=sparta)
        self._create_statement(human, mortal, Relation.from_string("is"), modifier=sparta)

        beliefs_by_key = self.belief_service._base_beliefs_for_agent(self.agent.id, athens, None)
        self.categorical_deduction_service.resolve_deductions(
            beliefs_by_key,
            modifier=athens,
            max_depth=5,
        )
        belief_key = self.belief_service._belief_key_from_values(
            agent_id=self.agent.id,
            subject_id=socrates.id,
            object_id=mortal.id,
            relation=Relation.from_string("is"),
            modifier=athens,
        )
        belief = beliefs_by_key.get(belief_key)
        self.assertIsNotNone(belief)
        deductions = self._deductions_from_belief(belief)
        self.assertEqual(len(deductions), 1)
        modifier = deductions[0].modifier
        self.assertIsNotNone(modifier)
        if modifier is None:
            self.fail("Expected deduction modifier to be set")
        self.assertEqual(modifier.location, "Athens")

    def test_no_deduction_with_location_mismatch_when_time_matches(self):
        socrates = self._create_notion("socrates", "socrates_id")
        human = self._create_notion("human", "human_id")
        mortal = self._create_notion("mortal", "mortal_id")

        today = Time(year=2024, month=1, day=1)
        athens = Modifier(location="Athens", from_time=today)
        sparta = Modifier(location="Sparta", from_time=today)

        self._create_statement(human, mortal, Relation.from_string("is"), modifier=athens)
        self._create_statement(socrates, human, Relation.from_string("is"), modifier=sparta)

        beliefs_by_key = self.belief_service._base_beliefs_for_agent(self.agent.id, athens, None)
        self.categorical_deduction_service.resolve_deductions(
            beliefs_by_key,
            modifier=athens,
            max_depth=5,
        )
        belief_key = self.belief_service._belief_key_from_values(
            agent_id=self.agent.id,
            subject_id=socrates.id,
            object_id=mortal.id,
            relation=Relation.from_string("is"),
            modifier=athens,
        )
        belief = beliefs_by_key.get(belief_key)
        self.assertIsNone(belief)

    def test_no_deduction_with_time_mismatch_when_location_matches(self):
        socrates = self._create_notion("socrates", "socrates_id")
        human = self._create_notion("human", "human_id")
        mortal = self._create_notion("mortal", "mortal_id")

        today = Modifier(location="Athens", from_time=Time(year=2024, month=1, day=1))
        tomorrow = Modifier(location="Athens", from_time=Time(year=2024, month=1, day=2))

        self._create_statement(human, mortal, Relation.from_string("is"), modifier=today)
        self._create_statement(socrates, human, Relation.from_string("is"), modifier=tomorrow)

        beliefs_by_key = self.belief_service._base_beliefs_for_agent(self.agent.id, today, None)
        self.categorical_deduction_service.resolve_deductions(
            beliefs_by_key,
            modifier=today,
            max_depth=5,
        )
        belief_key = self.belief_service._belief_key_from_values(
            agent_id=self.agent.id,
            subject_id=socrates.id,
            object_id=mortal.id,
            relation=Relation.from_string("is"),
            modifier=today,
        )
        belief = beliefs_by_key.get(belief_key)
        self.assertIsNone(belief)

    def test_is_a_preserves_custom_relation(self):
        socrates = self._create_notion("socrates", "socrates_id")
        human = self._create_notion("human", "human_id")
        wisdom = self._create_notion("wisdom", "wisdom_id")

        self._create_statement(human, wisdom, Relation.from_string("has property"))
        self._create_statement(socrates, human, Relation.from_string("is"))

        beliefs_by_key = self.belief_service._base_beliefs_for_agent(self.agent.id, None, None)
        self.categorical_deduction_service.resolve_deductions(
            beliefs_by_key,
            modifier=None,
            max_depth=5,
        )
        belief_key = self.belief_service._belief_key_from_values(
            agent_id=self.agent.id,
            subject_id=socrates.id,
            object_id=wisdom.id,
            relation=Relation.from_string("has property"),
            modifier=None,
        )
        belief = beliefs_by_key.get(belief_key)
        self.assertIsNotNone(belief)
        deductions = self._deductions_from_belief(belief)
        self.assertEqual(len(deductions), 1)

    def test_is_a_passes_through_custom_relation(self):
        socrates = self._create_notion("socrates", "socrates_id")
        human = self._create_notion("human", "human_id")
        property_notion = self._create_notion("property", "property_id")

        self._create_statement(human, property_notion, Relation.from_string("owns"))
        self._create_statement(socrates, human, Relation.from_string("is"))

        beliefs_by_key = self.belief_service._base_beliefs_for_agent(self.agent.id, None, None)
        self.categorical_deduction_service.resolve_deductions(
            beliefs_by_key,
            modifier=None,
            max_depth=5,
        )
        belief_key = self.belief_service._belief_key_from_values(
            agent_id=self.agent.id,
            subject_id=socrates.id,
            object_id=property_notion.id,
            relation=Relation.from_string("owns"),
            modifier=None,
        )
        belief = beliefs_by_key.get(belief_key)
        self.assertIsNotNone(belief)
        deductions = self._deductions_from_belief(belief)
        self.assertEqual(len(deductions), 1)

    def test_no_deduction_with_mismatched_time_modifiers(self):
        socrates = self._create_notion("socrates", "socrates_id")
        human = self._create_notion("human", "human_id")
        mortal = self._create_notion("mortal", "mortal_id")

        today = Modifier(from_time=Time(year=2024, month=1, day=1))
        tomorrow = Modifier(from_time=Time(year=2024, month=1, day=2))

        self._create_statement(human, mortal, Relation.from_string("is"), modifier=today)
        self._create_statement(socrates, human, Relation.from_string("is"), modifier=tomorrow)

        beliefs_by_key = self.belief_service._base_beliefs_for_agent(self.agent.id, today, None)
        self.categorical_deduction_service.resolve_deductions(
            beliefs_by_key,
            modifier=today,
            max_depth=5,
        )
        belief_key = self.belief_service._belief_key_from_values(
            agent_id=self.agent.id,
            subject_id=socrates.id,
            object_id=mortal.id,
            relation=Relation.from_string("is"),
            modifier=today,
        )
        belief = beliefs_by_key.get(belief_key)
        self.assertIsNone(belief)

    def test_no_deduction_with_minute_offset_modifiers(self):
        socrates = self._create_notion("socrates", "socrates_id")
        human = self._create_notion("human", "human_id")
        mortal = self._create_notion("mortal", "mortal_id")

        now = Modifier(from_time=Time(year=2024, month=1, day=1, hour=10, minute=0))
        in_two_minutes = Modifier(from_time=Time(year=2024, month=1, day=1, hour=10, minute=2))

        self._create_statement(human, mortal, Relation.from_string("is"), modifier=now)
        self._create_statement(socrates, human, Relation.from_string("is"), modifier=in_two_minutes)

        beliefs_by_key = self.belief_service._base_beliefs_for_agent(self.agent.id, now, None)
        self.categorical_deduction_service.resolve_deductions(
            beliefs_by_key,
            modifier=now,
            max_depth=5,
        )
        belief_key = self.belief_service._belief_key_from_values(
            agent_id=self.agent.id,
            subject_id=socrates.id,
            object_id=mortal.id,
            relation=Relation.from_string("is"),
            modifier=now,
        )
        belief = beliefs_by_key.get(belief_key)
        self.assertIsNone(belief)

    def test_deduction_with_matching_time_modifiers(self):
        socrates = self._create_notion("socrates", "socrates_id")
        human = self._create_notion("human", "human_id")
        mortal = self._create_notion("mortal", "mortal_id")

        today = Modifier(from_time=Time(year=2024, month=1, day=1))

        self._create_statement(human, mortal, Relation.from_string("is"), modifier=today)
        self._create_statement(socrates, human, Relation.from_string("is"), modifier=today)

        beliefs_by_key = self.belief_service._base_beliefs_for_agent(self.agent.id, today, None)
        self.categorical_deduction_service.resolve_deductions(
            beliefs_by_key,
            modifier=today,
            max_depth=5,
        )
        belief_key = self.belief_service._belief_key_from_values(
            agent_id=self.agent.id,
            subject_id=socrates.id,
            object_id=mortal.id,
            relation=Relation.from_string("is"),
            modifier=today,
        )
        belief = beliefs_by_key.get(belief_key)
        self.assertIsNotNone(belief)
        deductions = self._deductions_from_belief(belief)
        self.assertEqual(len(deductions), 1)
        modifier = deductions[0].modifier
        self.assertIsNotNone(modifier)
        if modifier is None:
            self.fail("Expected deduction modifier to be set")
        from_time = modifier.from_time
        self.assertIsNotNone(from_time)
        if from_time is None:
            self.fail("Expected deduction from_time to be set")
        self.assertEqual(from_time.day, 1)


if __name__ == "__main__":
    unittest.main()



from abc import ABC, abstractmethod
from .observable_service import ObservableService


class StatementObserver(ABC):
    @abstractmethod
    def on_statement_created(self, statement):
        pass

    @abstractmethod
    def on_statement_updated(self, statement):
        pass


class StatementObservableService(ObservableService):
    def notify_statement_created(self, statement) -> None:
        for observer in self._observers:
            try:
                observer.on_statement_created(statement)
            except (AttributeError, TypeError):
                pass

    def notify_statement_updated(self, statement) -> None:
        for observer in self._observers:
            try:
                observer.on_statement_updated(statement)
            except (AttributeError, TypeError):
                pass


from abc import ABC, abstractmethod
from typing import List, Dict, Any, Optional
from analogai.foundation.common.design_patterns.singleton import Singleton

class QueryStorageAdapter(ABC):
    
    @abstractmethod
    def store(self, collection: str, data: Any) -> None:
        pass

    @abstractmethod
    def retrieve_by_id(self, collection: str, id: str) -> Optional[Any]:
        pass

    @abstractmethod
    def retrieve_by_attributes(self, collection: str, query: Dict[str, Any]) -> List[Any]:
        pass

    @abstractmethod
    def delete(self, collection: str, id: str) -> bool:
        pass

    @abstractmethod
    def clear(self, collection: str) -> None:
        pass

    @abstractmethod
    def query(self, collection: str, query_dict: Dict[str, Any]) -> 'QueryBuilder':
        pass

    @abstractmethod
    def count(self, collection: str, query_dict: Optional[Dict[str, Any]] = None) -> int:
        pass

class QueryBuilder:
    
    def __init__(self, storage_adapter: QueryStorageAdapter, collection: str, query_dict: Dict[str, Any]):
        self.storage_adapter = storage_adapter
        self.collection = collection
        self.query_dict = query_dict
        self._limit_start: Optional[int] = None
        self._limit_count: Optional[int] = None
    
    def limit(self, start: int, count: int) -> 'QueryBuilder':
        self._limit_start = start
        self._limit_count = count
        return self
    
    def get(self) -> List[Any]:
        results = self.storage_adapter.retrieve_by_attributes(self.collection, self.query_dict)
        
        if self._limit_start is not None and self._limit_count is not None:
            results = results[self._limit_start:self._limit_start + self._limit_count]
        
        return results
    
    def count(self) -> int:
        return self.storage_adapter.count(self.collection, self.query_dict)


import asyncio
import unittest
from unittest.mock import Mock, patch
from fastapi import Request
from analogai.foundation.extensions.authentication.authentication_factory import AuthenticationFactory
from analogai.foundation.modules.user.user import User
from analogai.foundation.extensions.authentication.models import TokenPayload

class TestAuthenticationWorkflow(unittest.TestCase):
    def setUp(self):
        self.user_service = Mock()
        self.user = User(id="user-123", name="Test User", email="test@example.com")

    @patch("analogai.foundation.extensions.authentication.authentication_factory.create_repository_strategy")
    @patch("analogai.foundation.extensions.authentication.authentication_factory.Auth0JWTService")
    def test_authentication_dependency_returns_user(self, auth0_cls, repository_selector):
        repository_selector.return_value.get_user_repository.return_value = self.user_service
        jwt_service = auth0_cls.return_value
        jwt_service.authenticate_user.return_value = self.user

        factory = AuthenticationFactory()
        dependency = factory.create_authentication_dependency()

        request = Mock(spec=Request)
        request.headers = {"Authorization": "Bearer auth-token"}

        result_user = asyncio.run(dependency(request))

        jwt_service.authenticate_user.assert_called_once_with("auth-token")
        self.assertIs(result_user, self.user)

    @patch("analogai.foundation.extensions.authentication.authentication_factory.create_repository_strategy")
    @patch("analogai.foundation.extensions.authentication.authentication_factory.Auth0JWTService")
    def test_token_extraction_dependency_returns_payload(self, auth0_cls, repository_selector):
        repository_selector.return_value.get_user_repository.return_value = self.user_service
        expected_payload = TokenPayload(
            token="raw-token",
            email="payload@example.com",
            user_id="payload-user",
            name="Payload User",
            provider="auth0",
        )
        jwt_service = auth0_cls.return_value
        jwt_service.extract_payload.return_value = expected_payload

        factory = AuthenticationFactory()
        dependency = factory.create_token_extraction_dependency()

        request = Mock(spec=Request)
        request.headers = {"Authorization": "Bearer raw-token"}

        payload = asyncio.run(dependency(request))

        jwt_service.extract_payload.assert_called_once_with("raw-token")
        self.assertEqual(payload, expected_payload)

if __name__ == "__main__":
    unittest.main()




# import unittest

# from analogai.foundation.infrastructure.config import ENVIRONMENT

# class EnvironmentCheckMixin:
#     def setUp(self):
#         super().setUp()
#         if ENVIRONMENT == 'production':
#             self.skipTest(
#                 "CRITICAL: Tests cannot be run in production environment! "
#                 "Set ENVIRONMENT=staging or ENVIRONMENT=development to run tests."
#             )

# def prevent_tests_in_production():
#     if ENVIRONMENT == 'production':
#         raise RuntimeError(
#             "CRITICAL: Tests cannot be run in production environment! "
#             "Set ENVIRONMENT=staging or ENVIRONMENT=development to run tests."
#         )

# prevent_tests_in_production()

